from __future__ import annotations

from qtpy.QtCore import QPoint, QRect, QSize, Qt
from qtpy.QtWidgets import QLayout, QLayoutItem, QSizePolicy, QStyle, QWidget


class QFlowLayout(QLayout):
    """Layout that handles different window sizes.

    The widget placement changes depending on the width of the application window.

    Code translated from C++ at:
    <https://code.qt.io/cgit/qt/qtbase.git/tree/examples/widgets/layouts/flowlayout>

    described at: <https://doc.qt.io/qt-6/qtwidgets-layouts-flowlayout-example.html>

    See also: <https://doc.qt.io/qt-6/layout.html>

    Parameters
    ----------
    parent : QWidget, optional
        The parent widget, by default None
    """

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)

        self._item_list: list[QLayoutItem] = []
        self._h_space = -1
        self._v_space = -1

    def __del__(self) -> None:
        while item := self.takeAt(0):
            del item

    def addItem(self, item: QLayoutItem | None) -> None:
        """Add an item to the layout."""
        if item:
            self._item_list.append(item)

    def setHorizontalSpacing(self, space: int | None) -> None:
        """Set the horizontal spacing.

        If None or -1, the spacing is set to the default value based on the style
        of the parent widget.
        """
        self._h_space = -1 if space is None else space

    def horizontalSpacing(self) -> int:
        """Return the horizontal spacing."""
        if self._h_space >= 0:
            return self._h_space
        return self._smartSpacing(QStyle.PixelMetric.PM_LayoutHorizontalSpacing)

    def setVerticalSpacing(self, space: int | None) -> None:
        """Set the vertical spacing.

        If None or -1, the spacing is set to the default value based on the style
        of the parent widget.
        """
        self._v_space = -1 if space is None else space

    def verticalSpacing(self) -> int:
        """Return the vertical spacing."""
        if self._v_space >= 0:
            return self._v_space
        return self._smartSpacing(QStyle.PixelMetric.PM_LayoutVerticalSpacing)

    def expandingDirections(self) -> Qt.Orientation:
        """Return the expanding directions.

        These are the Qt::Orientations in which the layout can make use of more space
        than its sizeHint().
        """
        return Qt.Orientation.Horizontal

    def hasHeightForWidth(self) -> bool:
        """Return whether the layout handles height for width."""
        return True

    def heightForWidth(self, width: int) -> int:
        """Return the height for a given width.

        `heightForWidth()` passes the width on to _doLayout() which in turn uses the
        width as an argument for the layout rect, i.e., the bounds in which the items
        are laid out. This rect does not include the layout margin().
        """
        return self._doLayout(QRect(0, 0, width, 0), True)

    def count(self) -> int:
        """Return the number of items in the layout."""
        return len(self._item_list)

    def itemAt(self, index: int) -> QLayoutItem | None:
        """Return the item at the given index, or None if the index is out of range."""
        try:
            return self._item_list[index]
        except IndexError:
            return None

    def minimumSize(self) -> QSize:
        """Return the minimum size of the layout."""
        size = QSize()
        for item in self._item_list:
            size = size.expandedTo(item.minimumSize())

        margins = self.contentsMargins()
        size += QSize(
            margins.left() + margins.right(), margins.top() + margins.bottom()
        )
        return size

    def setGeometry(self, rect: QRect) -> None:
        """Set the geometry of the layout.

        This triggers a re-layout of the items.
        """
        super().setGeometry(rect)
        self._doLayout(rect)

    def sizeHint(self) -> QSize:
        """Return the size hint of the layout."""
        return self.minimumSize()

    def takeAt(self, index: int) -> QLayoutItem | None:
        """Remove and return the item at the given index.  Or return None."""
        if 0 <= index < len(self._item_list):
            return self._item_list.pop(index)
        return None

    def _doLayout(self, rect: QRect, test_only: bool = False) -> int:
        """Arrange the items in the layout.

        If test_only is True, the items are not actually laid out, but the height
        that the layout would have with the given width is returned.
        """
        left, top, right, bottom = self.getContentsMargins()
        effective_rect = rect.adjusted(left, top, -right, -bottom)  # type: ignore
        x = effective_rect.x()
        y = effective_rect.y()
        line_height = 0

        for item in self._item_list:
            if (wid := item.widget()) and (style := wid.style()):
                space_x = self.horizontalSpacing()
                space_y = self.verticalSpacing()
                if space_x == -1:
                    space_x = style.layoutSpacing(
                        QSizePolicy.ControlType.PushButton,
                        QSizePolicy.ControlType.PushButton,
                        Qt.Orientation.Horizontal,
                    )
                if space_y == -1:
                    space_y = style.layoutSpacing(
                        QSizePolicy.ControlType.PushButton,
                        QSizePolicy.ControlType.PushButton,
                        Qt.Orientation.Vertical,
                    )

                # next_x is the x-coordinate of the right edge of the item
                next_x = x + item.sizeHint().width() + space_x
                # if the item is not the first one in a line, add the spacing
                # to the left of it
                if next_x - space_x > effective_rect.right() and line_height > 0:
                    x = effective_rect.x()
                    y = y + line_height + space_y
                    next_x = x + item.sizeHint().width() + space_x
                    line_height = 0

                # if this is not a test run, move the item to its proper place
                if not test_only:
                    item.setGeometry(QRect(QPoint(x, y), item.sizeHint()))

                x = next_x
                line_height = max(line_height, item.sizeHint().height())

        return y + line_height - rect.y() + bottom

    def _smartSpacing(self, pm: QStyle.PixelMetric) -> int:
        """Return the smart spacing based on the style of the parent widget."""
        if isinstance(parent := self.parent(), QWidget) and (style := parent.style()):
            return style.pixelMetric(pm, None, parent)
        return -1



try:
    import cmap
except ImportError as e:
    raise ImportError(
        "The cmap package is required to use superqt colormap utilities. "
        "Install it with `pip install cmap` or `pip install superqt[cmap]`."
    ) from e
else:
    del cmap

from ._catalog_combo import CmapCatalogComboBox
from ._cmap_combo import QColormapComboBox
from ._cmap_item_delegate import QColormapItemDelegate
from ._cmap_line_edit import QColormapLineEdit
from ._cmap_utils import draw_colormap

__all__ = [
    "CmapCatalogComboBox",
    "QColormapComboBox",
    "QColormapItemDelegate",
    "QColormapLineEdit",
    "draw_colormap",
]



"""Adapted for python from the KDToolBox.

https://github.com/KDAB/KDToolBox/tree/master/qt/KDSignalThrottler

MIT License

Copyright (C) 2019-2022 Klarälvdalens Datakonsult AB, a KDAB Group company,
info@kdab.com

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""

from __future__ import annotations

import warnings
from concurrent.futures import Future
from contextlib import suppress
from enum import IntFlag, auto
from functools import wraps
from inspect import signature
from types import MethodType
from typing import TYPE_CHECKING, Any, Generic, TypeVar, overload
from weakref import WeakKeyDictionary, WeakMethod

from qtpy.QtCore import QObject, Qt, QTimer, Signal

from ._util import get_max_args

if TYPE_CHECKING:
    from collections.abc import Callable

    from typing_extensions import ParamSpec

    P = ParamSpec("P")
# maintain runtime compatibility with older typing_extensions
else:
    try:
        from typing_extensions import ParamSpec

        P = ParamSpec("P")
    except ImportError:
        P = TypeVar("P")

R = TypeVar("R")
REF_ERROR = (
    "To use qthrottled or qdebounced as a method decorator, "
    "objects must have  `__dict__` or be weak referenceable. "
    "Please either add `__weakref__` to `__slots__` or use"
    "qthrottled/qdebounced as a function (not a decorator)."
)


class Kind(IntFlag):
    Throttler = auto()
    Debouncer = auto()


class EmissionPolicy(IntFlag):
    Trailing = auto()
    Leading = auto()


class GenericSignalThrottler(QObject):
    triggered = Signal()
    timeoutChanged = Signal(int)
    timerTypeChanged = Signal(Qt.TimerType)

    def __init__(
        self,
        kind: Kind,
        emissionPolicy: EmissionPolicy,
        parent: QObject | None = None,
    ) -> None:
        super().__init__(parent)

        self._kind = kind
        self._emissionPolicy = emissionPolicy
        self._hasPendingEmission = False

        self._timer = QTimer(parent=self)
        self._timer.setSingleShot(True)
        self._timer.setTimerType(Qt.TimerType.PreciseTimer)
        self._timer.timeout.connect(self._maybeEmitTriggered)

    def kind(self) -> Kind:
        """Return the kind of throttler (throttler or debouncer)."""
        return self._kind

    def emissionPolicy(self) -> EmissionPolicy:
        """Return the emission policy (trailing or leading)."""
        return self._emissionPolicy

    def timeout(self) -> int:
        """Return current timeout in milliseconds."""
        return self._timer.interval()

    def setTimeout(self, timeout: int) -> None:
        """Set timeout in milliseconds."""
        if self._timer.interval() != timeout:
            self._timer.setInterval(timeout)
            self.timeoutChanged.emit(timeout)

    def timerType(self) -> Qt.TimerType:
        """Return current `Qt.TimerType`."""
        return self._timer.timerType()

    def setTimerType(self, timerType: Qt.TimerType) -> None:
        """Set current Qt.TimerType."""
        if self._timer.timerType() != timerType:
            self._timer.setTimerType(timerType)
            self.timerTypeChanged.emit(timerType)

    def throttle(self) -> None:
        """Emit triggered if not running, then start timer."""
        # public slot
        self._hasPendingEmission = True
        # Emit only if we haven't emitted already. We know if that's
        # the case by checking if the timer is running.
        if (
            self._emissionPolicy is EmissionPolicy.Leading
            and not self._timer.isActive()
        ):
            self._emitTriggered()

        # The timer is started in all cases. If we got a signal, and we're Leading,
        # and we did emit because of that, then we don't re-emit when the timer fires
        # (unless we get ANOTHER signal).
        if self._kind is Kind.Throttler:  # sourcery skip: merge-duplicate-blocks
            if not self._timer.isActive():
                self._timer.start()  # actual start, not restart
        elif self._kind is Kind.Debouncer:
            self._timer.start()  # restart

    def cancel(self) -> None:
        """Cancel any pending emissions."""
        self._hasPendingEmission = False

    def flush(self, restart_timer: bool = True) -> None:
        """
        Force emission of any pending emissions.

        Parameters
        ----------
        restart_timer : bool
            Whether to restart the timer after flushing.
            Defaults to True.
        """
        self._maybeEmitTriggered(restart_timer=restart_timer)

    def _emitTriggered(self) -> None:
        self._hasPendingEmission = False
        self.triggered.emit()
        self._timer.start()

    def _maybeEmitTriggered(self, restart_timer: bool = True) -> None:
        if self._hasPendingEmission:
            self._emitTriggered()
        if not restart_timer:
            self._timer.stop()

    Kind = Kind
    EmissionPolicy = EmissionPolicy


# ### Convenience classes ###


class QSignalThrottler(GenericSignalThrottler):
    """A Signal Throttler.

    This object's `triggered` signal will emit at most once per timeout
    (set with setTimeout()).
    """

    def __init__(
        self,
        policy: EmissionPolicy = EmissionPolicy.Leading,
        parent: QObject | None = None,
    ) -> None:
        super().__init__(Kind.Throttler, policy, parent)


class QSignalDebouncer(GenericSignalThrottler):
    """A Signal Debouncer.

    This object's `triggered` signal will not be emitted until `self.timeout()`
    milliseconds have elapsed since the last time `triggered` was emitted.
    """

    def __init__(
        self,
        policy: EmissionPolicy = EmissionPolicy.Trailing,
        parent: QObject | None = None,
    ) -> None:
        super().__init__(Kind.Debouncer, policy, parent)


# below here part is unique to superqt (not from KD)


def _weak_func(func: Callable[P, R]) -> Callable[P, R]:
    if isinstance(func, MethodType):
        # this is a bound method, we need to avoid strong references
        try:
            weak_method = WeakMethod(func)
        except TypeError as e:
            raise TypeError(REF_ERROR) from e

        def weak_func(*args, **kwargs):
            if method := weak_method():
                return method(*args, **kwargs)
            warnings.warn(
                "Method has been garbage collected", RuntimeWarning, stacklevel=2
            )

        return weak_func

    return func


class ThrottledCallable(GenericSignalThrottler, Generic[P, R]):
    def __init__(
        self,
        func: Callable[P, R],
        kind: Kind,
        emissionPolicy: EmissionPolicy,
        parent: QObject | None = None,
    ) -> None:
        super().__init__(kind, emissionPolicy, parent)

        self._future: Future[R] = Future()

        self._is_static_method: bool = False
        if isinstance(func, staticmethod):
            self._is_static_method = True
            func = func.__func__

        max_args = get_max_args(func)
        with suppress(TypeError, ValueError):
            self.__signature__ = signature(func)

        self._func = _weak_func(func)
        self.__wrapped__ = self._func

        self._args: tuple = ()
        self._kwargs: dict = {}
        self.triggered.connect(self._set_future_result)
        self._name = None

        self._obj_dkt: WeakKeyDictionary[Any, ThrottledCallable] = WeakKeyDictionary()

        # even if we were to compile __call__ with a signature matching that of func,
        # PySide wouldn't correctly inspect the signature of the ThrottledCallable
        # instance: https://bugreports.qt.io/browse/PYSIDE-2423
        # so we do it ourselves and limit the number of positional arguments
        # that we pass to func
        self._max_args: int | None = max_args

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> "Future[R]":  # noqa
        if not self._future.done():
            self._future.cancel()

        self._future = Future()
        self._args = args
        self._kwargs = kwargs

        self.throttle()
        return self._future

    def _set_future_result(self):
        result = self._func(*self._args[: self._max_args], **self._kwargs)
        self._future.set_result(result)

    def __set_name__(self, owner, name):
        if not self._is_static_method:
            self._name = name

    def _get_throttler(self, instance, owner, parent, obj, name):
        try:
            bound_method = self._func.__get__(instance, owner)
        except Exception as e:  # pragma: no cover
            raise RuntimeError(
                f"Failed to bind function {self._func!r} to object {instance!r}"
            ) from e
        throttler = ThrottledCallable(
            bound_method,
            self._kind,
            self._emissionPolicy,
            parent=parent,
        )
        throttler.setTimerType(self.timerType())
        throttler.setTimeout(self.timeout())
        try:
            setattr(obj, name, throttler)
        except AttributeError:
            try:
                self._obj_dkt[obj] = throttler
            except TypeError as e:
                raise TypeError(REF_ERROR) from e
        return throttler

    def __get__(self, instance, owner):
        if instance is None or not self._name:
            return self

        if instance in self._obj_dkt:
            return self._obj_dkt[instance]

        parent = self.parent()
        if parent is None and isinstance(instance, QObject):
            parent = instance

        return self._get_throttler(instance, owner, parent, instance, self._name)


@overload
def qthrottled(
    func: Callable[P, R],
    timeout: int = 100,
    leading: bool = True,
    timer_type: Qt.TimerType = Qt.TimerType.PreciseTimer,
    parent: QObject | None = None,
) -> ThrottledCallable[P, R]: ...


@overload
def qthrottled(
    func: None = ...,
    timeout: int = 100,
    leading: bool = True,
    timer_type: Qt.TimerType = Qt.TimerType.PreciseTimer,
    parent: QObject | None = None,
) -> Callable[[Callable[P, R]], ThrottledCallable[P, R]]: ...


def qthrottled(
    func: Callable[P, R] | None = None,
    timeout: int = 100,
    leading: bool = True,
    timer_type: Qt.TimerType = Qt.TimerType.PreciseTimer,
    parent: QObject | None = None,
) -> ThrottledCallable[P, R] | Callable[[Callable[P, R]], ThrottledCallable[P, R]]:
    """Creates a throttled function that invokes func at most once per timeout.

    The throttled function comes with a `cancel` method to cancel delayed func
    invocations and a `flush` method to immediately invoke them. Options
    to indicate whether func should be invoked on the leading and/or trailing
    edge of the wait timeout. The func is invoked with the last arguments provided
    to the throttled function. Subsequent calls to the throttled function return
    the result of the last func invocation.

    This decorator may be used with or without parameters.

    Parameters
    ----------
    func : Callable
        A function to throttle
    timeout : int
        Timeout in milliseconds to wait before allowing another call, by default 100
    leading : bool
        Whether to invoke the function on the leading edge of the wait timer,
        by default True
    timer_type : Qt.TimerType
        The timer type. by default `Qt.TimerType.PreciseTimer`
        One of:
            - `Qt.PreciseTimer`: Precise timers try to keep millisecond accuracy
            - `Qt.CoarseTimer`: Coarse timers try to keep accuracy within 5% of the
              desired interval
            - `Qt.VeryCoarseTimer`: Very coarse timers only keep full second accuracy
    parent: QObject or None
        Parent object for timer. If using qthrottled as function it may be useful
        for cleaning data
    """
    return _make_decorator(func, timeout, leading, timer_type, Kind.Throttler, parent)


@overload
def qdebounced(
    func: Callable[P, R],
    timeout: int = 100,
    leading: bool = False,
    timer_type: Qt.TimerType = Qt.TimerType.PreciseTimer,
    parent: QObject | None = None,
) -> ThrottledCallable[P, R]: ...


@overload
def qdebounced(
    func: None = ...,
    timeout: int = 100,
    leading: bool = False,
    timer_type: Qt.TimerType = Qt.TimerType.PreciseTimer,
    parent: QObject | None = None,
) -> Callable[[Callable[P, R]], ThrottledCallable[P, R]]: ...


def qdebounced(
    func: Callable[P, R] | None = None,
    timeout: int = 100,
    leading: bool = False,
    timer_type: Qt.TimerType = Qt.TimerType.PreciseTimer,
    parent: QObject | None = None,
) -> ThrottledCallable[P, R] | Callable[[Callable[P, R]], ThrottledCallable[P, R]]:
    """Creates a debounced function that delays invoking `func`.

    `func` will not be invoked until `timeout` ms have elapsed since the last time
    the debounced function was invoked.

    The debounced function comes with a `cancel` method to cancel delayed func
    invocations and a `flush` method to immediately invoke them. Options
    indicate whether func should be invoked on the leading and/or trailing edge
    of the wait timeout. The func is invoked with the *last* arguments provided to
    the debounced function. Subsequent calls to the debounced function return the
    result of the last `func` invocation.

    This decorator may be used with or without parameters.

    Parameters
    ----------
    func : Callable
        A function to throttle
    timeout : int
        Timeout in milliseconds to wait before allowing another call, by default 100
    leading : bool
        Whether to invoke the function on the leading edge of the wait timer,
        by default False
    timer_type : Qt.TimerType
        The timer type. by default `Qt.TimerType.PreciseTimer`
        One of:
            - `Qt.PreciseTimer`: Precise timers try to keep millisecond accuracy
            - `Qt.CoarseTimer`: Coarse timers try to keep accuracy within 5% of the
              desired interval
            - `Qt.VeryCoarseTimer`: Very coarse timers only keep full second accuracy
    parent: QObject or None
        Parent object for timer. If using qthrottled as function it may be useful
        for cleaning data
    """
    return _make_decorator(func, timeout, leading, timer_type, Kind.Debouncer, parent)


def _make_decorator(
    func: Callable[P, R] | None,
    timeout: int,
    leading: bool,
    timer_type: Qt.TimerType,
    kind: Kind,
    parent: QObject | None = None,
) -> ThrottledCallable[P, R] | Callable[[Callable[P, R]], ThrottledCallable[P, R]]:
    def deco(func: Callable[P, R]) -> ThrottledCallable[P, R]:
        nonlocal parent

        instance: object | None = getattr(func, "__self__", None)
        if isinstance(instance, QObject) and parent is None:
            parent = instance
        policy = EmissionPolicy.Leading if leading else EmissionPolicy.Trailing
        obj = ThrottledCallable(func, kind, policy, parent=parent)
        obj.setTimerType(timer_type)
        obj.setTimeout(timeout)

        if instance is not None:
            # this is a bound method, we need to avoid strong references,
            # and functools.wraps will prevent garbage collection on bound methods
            return obj
        return wraps(func)(obj)

    return deco(func) if func is not None else deco



from __future__ import annotations

from typing import TYPE_CHECKING

from cmap import Colormap
from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import QComboBox, QCompleter, QWidget

from ._cmap_item_delegate import QColormapItemDelegate
from ._cmap_line_edit import QColormapLineEdit
from ._cmap_utils import try_cast_colormap

if TYPE_CHECKING:
    from collections.abc import Container

    from cmap._catalog import Category, Interpolation
    from qtpy.QtGui import QKeyEvent


class CmapCatalogComboBox(QComboBox):
    """A combo box for selecting a colormap from the entire cmap catalog.

    Parameters
    ----------
    parent : QWidget, optional
        The parent widget.
    prefer_short_names : bool, optional
        If True (default), short names (without the namespace prefix) will be
        preferred over fully qualified names. In cases where the same short name is
        used in multiple namespaces, they will *all* be referred to by their fully
        qualified (namespaced) name.
    categories : Container[Category], optional
        If provided, only return names from the given categories.
    interpolation : Interpolation, optional
        If provided, only return names that have the given interpolation method.
    """

    currentColormapChanged = Signal(Colormap)

    def __init__(
        self,
        parent: QWidget | None = None,
        *,
        categories: Container[Category] = (),
        prefer_short_names: bool = True,
        interpolation: Interpolation | None = None,
    ) -> None:
        super().__init__(parent)

        # get valid names according to preferences
        word_list = sorted(
            Colormap.catalog().unique_keys(
                prefer_short_names=prefer_short_names,
                categories=categories,
                interpolation=interpolation,
            )
        )

        # initialize the combobox
        self.addItems(word_list)
        self.setInsertPolicy(QComboBox.InsertPolicy.NoInsert)
        self.setEditable(True)
        self.setDuplicatesEnabled(False)
        # (must come before setCompleter)
        self.setLineEdit(QColormapLineEdit(self))

        # setup the completer
        completer = QCompleter(word_list)
        completer.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        completer.setCompletionMode(QCompleter.CompletionMode.PopupCompletion)
        completer.setFilterMode(Qt.MatchFlag.MatchContains)
        completer.setModel(self.model())
        self.setCompleter(completer)

        # set the delegate for both the popup and the combobox
        delegate = QColormapItemDelegate()
        if popup := completer.popup():
            popup.setItemDelegate(delegate)
        self.setItemDelegate(delegate)

        self.currentTextChanged.connect(self._on_text_changed)

    def currentColormap(self) -> Colormap | None:
        """Returns the currently selected Colormap or None if not yet selected."""
        return try_cast_colormap(self.currentText())

    def keyPressEvent(self, e: QKeyEvent | None) -> None:
        if e and e.key() in (Qt.Key.Key_Enter, Qt.Key.Key_Return):
            # select the first completion when pressing enter if the popup is visible
            if (completer := self.completer()) and completer.completionCount():
                self.lineEdit().setText(completer.currentCompletion())  # type: ignore
        return super().keyPressEvent(e)

    def _on_text_changed(self, text: str) -> None:
        if (cmap := try_cast_colormap(text)) is not None:
            self.currentColormapChanged.emit(cmap)



import sys
from typing import TYPE_CHECKING

from qtpy.QtGui import QImage

if TYPE_CHECKING:
    import numpy as np


def qimage_to_array(img: QImage) -> "np.ndarray":
    """Convert QImage to an array.

    Parameters
    ----------
    img : QImage
        QImage to be converted.

    Returns
    -------
    arr : np.ndarray
        Numpy array of type uint8 and shape (h, w, 4). Index [0, 0] is the
        upper-left corner of the rendered region.
    """
    import numpy as np

    # cast to ARGB32 if necessary
    if img.format() != QImage.Format.Format_ARGB32:
        img = img.convertToFormat(QImage.Format.Format_ARGB32)

    h, w, c = img.height(), img.width(), 4

    # pyside returns a memoryview, pyqt returns a sizeless void pointer
    b = img.constBits()  # Returns a pointer to the first pixel data.
    if hasattr(b, "setsize"):
        b.setsize(h * w * c)

    # reshape to h, w, c
    arr = np.frombuffer(b, np.uint8).reshape(h, w, c)

    # reverse channel colors for numpy
    # On big endian we need to specify a different order
    if sys.byteorder == "big":
        return arr.take([1, 2, 3, 0], axis=2)  # pragma: no cover
    else:
        return arr.take([2, 1, 0, 3], axis=2)



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from qtpy.QtCore import QModelIndex, QObject, QPersistentModelIndex, QRect, QSize, Qt
from qtpy.QtGui import QColor, QPainter
from qtpy.QtWidgets import QStyle, QStyledItemDelegate, QStyleOptionViewItem

from ._cmap_utils import CMAP_ROLE, draw_colormap, pick_font_color, try_cast_colormap

if TYPE_CHECKING:
    from cmap import Colormap

DEFAULT_SIZE = QSize(80, 22)
DEFAULT_BORDER_COLOR = QColor(Qt.GlobalColor.transparent)


class QColormapItemDelegate(QStyledItemDelegate):
    """Delegate that draws colormaps into a QAbstractItemView item.

    Parameters
    ----------
    parent : QObject, optional
        The parent object.
    item_size : QSize, optional
        The size hint for each item, by default QSize(80, 22).
    fractional_colormap_width : float, optional
        The fraction of the widget width to use for the colormap swatch. If the
        colormap is full width (greater than 0.75), the swatch will be drawn behind
        the text. Otherwise, the swatch will be drawn to the left of the text.
        Default is 0.33.
    padding : int, optional
        The padding (in pixels) around the edge of the item, by default 1.
    checkerboard_size : int, optional
        Size (in pixels) of the checkerboard pattern to draw behind colormaps with
        transparency, by default 4. If 0, no checkerboard is drawn.
    """

    def __init__(
        self,
        parent: QObject | None = None,
        *,
        item_size: QSize = DEFAULT_SIZE,
        fractional_colormap_width: float = 1,
        padding: int = 1,
        checkerboard_size: int = 4,
    ) -> None:
        super().__init__(parent)
        self._item_size = item_size
        self._colormap_fraction = fractional_colormap_width
        self._padding = padding
        self._border_color: QColor | None = DEFAULT_BORDER_COLOR
        self._checkerboard_size = checkerboard_size

    def sizeHint(
        self, option: QStyleOptionViewItem, index: QModelIndex | QPersistentModelIndex
    ) -> QSize:
        return super().sizeHint(option, index).expandedTo(self._item_size)

    def paint(
        self,
        painter: QPainter,
        option: QStyleOptionViewItem,
        index: QModelIndex | QPersistentModelIndex,
    ) -> None:
        self.initStyleOption(option, index)
        rect = cast("QRect", option.rect)  # type: ignore
        selected = option.state & QStyle.StateFlag.State_Selected  # type: ignore
        text = index.data(Qt.ItemDataRole.DisplayRole)
        colormap: Colormap | None = index.data(CMAP_ROLE) or try_cast_colormap(text)

        if not colormap:  # pragma: no cover
            return super().paint(painter, option, index)

        painter.save()
        rect.adjust(self._padding, self._padding, -self._padding, -self._padding)
        cmap_rect = QRect(rect)
        cmap_rect.setWidth(int(rect.width() * self._colormap_fraction))

        lighter = 110 if selected else 100
        border = self._border_color if selected else None
        draw_colormap(
            painter,
            colormap,
            cmap_rect,
            lighter=lighter,
            border_color=border,
            checkerboard_size=self._checkerboard_size,
        )

        # # make new rect with the remaining space
        text_rect = QRect(rect)

        if self._colormap_fraction > 0.75:
            text_align = Qt.AlignmentFlag.AlignCenter
            alpha = 230 if selected else 140
            text_color = pick_font_color(colormap, alpha=alpha)
        else:
            text_align = Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter
            text_color = QColor(Qt.GlobalColor.black)
            text_rect.adjust(
                cmap_rect.width() + self._padding + 4, 0, -self._padding - 2, 0
            )

        painter.setPen(text_color)
        # cast to int works all the way back to Qt 5.12...
        # but the enum only works since Qt 5.14
        painter.drawText(text_rect, int(text_align), text)
        painter.restore()



from ._labeled import (
    QLabeledDoubleRangeSlider,
    QLabeledDoubleSlider,
    QLabeledRangeSlider,
    QLabeledSlider,
)
from ._range_style import MONTEREY_SLIDER_STYLES_FIX
from ._sliders import QDoubleRangeSlider, QDoubleSlider, QRangeSlider

__all__ = [
    "MONTEREY_SLIDER_STYLES_FIX",
    "QDoubleRangeSlider",
    "QDoubleSlider",
    "QLabeledDoubleRangeSlider",
    "QLabeledDoubleSlider",
    "QLabeledRangeSlider",
    "QLabeledSlider",
    "QRangeSlider",
]



from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from superqt.cmap import draw_colormap

__all__ = (
    "CodeSyntaxHighlight",
    "FunctionWorker",
    "GeneratorWorker",
    "QFlowLayout",
    "QMessageHandler",
    "QSignalDebouncer",
    "QSignalThrottler",
    "WorkerBase",
    "create_worker",
    "draw_colormap",
    "ensure_main_thread",
    "ensure_object_thread",
    "exceptions_as_dialog",
    "new_worker_qthread",
    "qdebounced",
    "qimage_to_array",
    "qthrottled",
    "signals_blocked",
    "thread_worker",
)

from ._code_syntax_highlight import CodeSyntaxHighlight
from ._ensure_thread import ensure_main_thread, ensure_object_thread
from ._errormsg_context import exceptions_as_dialog
from ._flow_layout import QFlowLayout
from ._img_utils import qimage_to_array
from ._message_handler import QMessageHandler
from ._misc import signals_blocked
from ._qthreading import (
    FunctionWorker,
    GeneratorWorker,
    WorkerBase,
    create_worker,
    new_worker_qthread,
    thread_worker,
)
from ._throttler import QSignalDebouncer, QSignalThrottler, qdebounced, qthrottled


def __getattr__(name: str) -> Any:  # pragma: no cover
    if name == "draw_colormap":
        from superqt.cmap import draw_colormap

        return draw_colormap
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



from qtpy import QT_VERSION
from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import QComboBox, QCompleter, QWidget

try:
    is_qt_bellow_5_14 = tuple(int(x) for x in QT_VERSION.split(".")[:2]) < (5, 14)
except ValueError:
    is_qt_bellow_5_14 = False


class QSearchableComboBox(QComboBox):
    """ComboCox with completer for fast search in multiple options."""

    if is_qt_bellow_5_14:
        textActivated = Signal(str)  # pragma: no cover

    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent)
        self.setEditable(True)
        self.completer_object = QCompleter()
        self.completer_object.setCaseSensitivity(Qt.CaseInsensitive)
        self.completer_object.setCompletionMode(QCompleter.PopupCompletion)
        self.completer_object.setFilterMode(Qt.MatchContains)
        self.setCompleter(self.completer_object)
        self.setInsertPolicy(QComboBox.NoInsert)
        if is_qt_bellow_5_14:  # pragma: no cover
            self.currentIndexChanged.connect(self._text_activated)

    def _text_activated(self):  # pragma: no cover
        self.textActivated.emit(self.currentText())

    def addItem(self, *args):
        super().addItem(*args)
        self.completer_object.setModel(self.model())

    def addItems(self, *args):
        super().addItems(*args)
        self.completer_object.setModel(self.model())

    def insertItem(self, *args) -> None:
        super().insertItem(*args)
        self.completer_object.setModel(self.model())

    def insertItems(self, *args) -> None:
        super().insertItems(*args)
        self.completer_object.setModel(self.model())



from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING, Any

from cmap import Colormap
from qtpy.QtCore import QPointF, QRect, QRectF, Qt
from qtpy.QtGui import QColor, QLinearGradient, QPaintDevice, QPainter

if TYPE_CHECKING:
    from cmap._colormap import ColorStopsLike

CMAP_ROLE = Qt.ItemDataRole.UserRole + 1


def draw_colormap(
    painter_or_device: QPainter | QPaintDevice,
    cmap: Colormap | ColorStopsLike,
    rect: QRect | QRectF | None = None,
    border_color: QColor | str | None = None,
    border_width: int = 1,
    lighter: int = 100,
    checkerboard_size: int = 4,
) -> None:
    """Draw a colormap onto a QPainter or QPaintDevice.

    Parameters
    ----------
    painter_or_device : QPainter | QPaintDevice
        A `QPainter` instance or a `QPaintDevice` (e.g. a QWidget or QPixmap) onto
        which to paint the colormap.
    cmap : Colormap | Any
        `cmap.Colormap` instance, or anything that can be converted to one (such as a
        string name of a colormap in the `cmap` catalog).
        https://cmap-docs.readthedocs.io/en/latest/colormaps/#colormaplike-objects
    rect : QRect | QRectF | None, optional
        A rect onto which to draw. If `None`, the `painter.viewport()` will be
        used.  by default `None`
    border_color : QColor | str | None
        If not `None`, a border of color `border_color` and width `border_width` is
        included around the edge, by default None.
    border_width : int, optional
        The width of the border to draw (provided `border_color` is not `None`),
        by default 2
    lighter : int, optional
        Percentage by which to lighten (or darken) the colors. Greater than 100
        lightens, less than 100 darkens, by default 100 (i.e. no change).
    checkerboard_size : bool, optional
        Size (in pixels) of the checkerboard pattern to draw, by default 5.
        If 0, no checkerboard is drawn.

    Examples
    --------
    ```python
    from qtpy.QtGui import QPixmap
    from qtpy.QtWidgets import QWidget
    from superqt.utils import draw_colormap

    viridis = "viridis"  # or cmap.Colormap('viridis')


    class W(QWidget):
        def paintEvent(self, event) -> None:
            draw_colormap(self, viridis, event.rect())


    # or draw onto a QPixmap
    pm = QPixmap(200, 200)
    draw_colormap(pm, viridis)
    ```
    """
    if isinstance(painter_or_device, QPainter):
        painter = painter_or_device
    elif isinstance(painter_or_device, QPaintDevice):
        painter = QPainter(painter_or_device)
    else:
        raise TypeError(
            "Expected a QPainter or QPaintDevice instance, "
            f"got {type(painter_or_device)!r} instead."
        )

    if (cmap_ := try_cast_colormap(cmap)) is None:
        raise TypeError(
            f"Expected a Colormap instance or something that can be "
            f"converted to one, got {cmap!r} instead."
        )

    if rect is None:
        rect = painter.viewport()

    painter.setPen(Qt.PenStyle.NoPen)

    if border_width and border_color is not None:
        # draw rect, and then contract it by border_width
        painter.setPen(QColor(border_color))
        painter.setBrush(Qt.BrushStyle.NoBrush)
        painter.drawRect(rect)
        rect = rect.adjusted(border_width, border_width, -border_width, -border_width)

    if checkerboard_size:
        _draw_checkerboard(painter, rect, checkerboard_size)

    if (
        cmap_.interpolation == "nearest"
        or getattr(cmap_.color_stops, "_interpolation", "") == "nearest"
    ):
        # XXX: this is a little bit of a hack.
        # when the interpolation is nearest, the last stop is often at 1.0
        # which means that the last color is not drawn.
        # to fix this, we shrink the drawing area slightly
        # it might not work well with unenvenly-spaced stops
        # (but those are uncommon for categorical colormaps)
        width = rect.width() - rect.width() / len(cmap_.color_stops)
        for stop in cmap_.color_stops:
            painter.setBrush(QColor(stop.color.hex).lighter(lighter))
            painter.drawRect(rect.adjusted(int(stop.position * width), 0, 0, 0))
    else:
        gradient = QLinearGradient(QPointF(rect.topLeft()), QPointF(rect.topRight()))
        for stop in cmap_.color_stops:
            gradient.setColorAt(stop.position, QColor(stop.color.hex).lighter(lighter))
        painter.setBrush(gradient)
        painter.drawRect(rect)

    # If we created a new Painter, free its resources
    if isinstance(painter_or_device, QPaintDevice):
        painter.end()


def _draw_checkerboard(
    painter: QPainter, rect: QRect | QRectF, checker_size: int
) -> None:
    darkgray = QColor("#969696")
    lightgray = QColor("#C8C8C8")
    sz = checker_size
    h, w = rect.height(), rect.width()
    left, top = rect.left(), rect.top()
    full_rows = h // sz
    full_cols = w // sz
    for row in range(int(full_rows) + 1):
        szh = sz if row < full_rows else int(h % sz)
        for col in range(int(full_cols) + 1):
            szw = sz if col < full_cols else int(w % sz)
            color = lightgray if (row + col) % 2 == 0 else darkgray
            painter.fillRect(int(col * sz + left), int(row * sz + top), szw, szh, color)


def try_cast_colormap(val: Any) -> Colormap | None:
    """Try to cast `val` to a Colormap instance, return None if it fails."""
    if isinstance(val, Colormap):
        return val
    with suppress(Exception):
        return Colormap(val)
    return None


def pick_font_color(cmap: Colormap, at_stop: float = 0.49, alpha: int = 255) -> QColor:
    """Pick a font shade that contrasts with the given colormap at `at_stop`."""
    if _is_dark(cmap, at_stop):
        return QColor(0, 0, 0, alpha)
    else:
        return QColor(255, 255, 255, alpha)


def _is_dark(cmap: Colormap, at_stop: float, threshold: float = 110) -> bool:
    """Return True if the color at `at_stop` is dark according to `threshold`."""
    color = cmap(at_stop)
    r, g, b, _a = color.rgba8
    return (r * 0.299 + g * 0.587 + b * 0.114) > threshold



from typing import TYPE_CHECKING, TypeAlias, Union

try:
    from pint import Quantity, Unit, UnitRegistry
    from pint.util import UnitsContainer
except ImportError as e:
    raise ImportError(
        "pint is required to use QQuantity.  Install it with `pip install pint`"
    ) from e

from qtpy.QtCore import Signal
from qtpy.QtWidgets import QComboBox, QDoubleSpinBox, QHBoxLayout, QSizePolicy, QWidget

from superqt.utils import signals_blocked

if TYPE_CHECKING:
    from decimal import Decimal


Number: TypeAlias = Union[int, float, "Decimal"]
UREG = UnitRegistry()
NULL_OPTION = "-----"
QOVERFLOW = 2**30
SI_BASES = {
    "[length]": "meter",
    "[time]": "second",
    "[current]": "ampere",
    "[luminosity]": "candela",
    "[mass]": "gram",
    "[substance]": "mole",
    "[temperature]": "kelvin",
}
DEFAULT_OPTIONS = {
    "[length]": ["km", "m", "mm", "µm"],
    "[time]": ["day", "hour", "min", "sec", "ms"],
    "[current]": ["A", "mA", "µA"],
    "[luminosity]": ["kcd", "cd", "mcd"],
    "[mass]": ["kg", "g", "mg", "µg"],
    "[substance]": ["mol", "mmol", "µmol"],
    "[temperature]": ["°C", "°F", "°K"],
    "radian": ["rad", "deg"],
}


class QQuantity(QWidget):
    """A combination QDoubleSpinBox and QComboBox for entering quantities.

    For this widget, `value()` returns a `pint.Quantity` object, while `setValue()`
    accepts either a number, `pint.Quantity`, a string that can be parsed by `pint`.

    Parameters
    ----------
    value : Union[str, pint.Quantity, Number]
        The initial value to display.  If a string, it will be parsed by `pint`.
    units : Union[pint.util.UnitsContainer, str, pint.Quantity], optional
        The units to use if `value` is a number.  If a string, it will be parsed by
        `pint`.  If a `pint.Quantity`, the units will be extracted from it.
    ureg : pint.UnitRegistry, optional
        The unit registry to use.  If not provided, the registry will be extracted
        from `value` if it is a `pint.Quantity`, otherwise the default registry will
        be used.
    parent : QWidget, optional
        The parent widget, by default None
    """

    valueChanged = Signal(Quantity)
    unitsChanged = Signal(Unit)
    dimensionalityChanged = Signal(UnitsContainer)

    def __init__(
        self,
        value: str | Quantity | Number = 0,
        units: UnitsContainer | str | Quantity | None = None,
        ureg: UnitRegistry | None = None,
        parent: QWidget | None = None,
    ) -> None:
        super().__init__(parent=parent)
        if ureg is None:
            ureg = value._REGISTRY if isinstance(value, Quantity) else UREG
        else:
            if not isinstance(ureg, UnitRegistry):
                raise TypeError(
                    f"ureg must be a pint.UnitRegistry, not {type(ureg).__name__}"
                )

        self._ureg = ureg
        self._value: Quantity = self._ureg.Quantity(value, units=units)

        # whether to preserve quantity equality when changing units or magnitude
        self._preserve_quantity: bool = False
        self._abbreviate_units: bool = True  # TODO: implement

        self._mag_spinbox = QDoubleSpinBox()
        self._mag_spinbox.setDecimals(3)
        self._mag_spinbox.setRange(-QOVERFLOW, QOVERFLOW - 1)
        self._mag_spinbox.setValue(float(self._value.magnitude))
        self._mag_spinbox.valueChanged.connect(self.setMagnitude)

        self._units_combo = QComboBox()
        self._units_combo.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self._units_combo.currentTextChanged.connect(self.setUnits)
        self._update_units_combo_choices()

        self.setLayout(QHBoxLayout())
        self.layout().addWidget(self._mag_spinbox)
        self.layout().addWidget(self._units_combo)
        self.layout().setContentsMargins(6, 0, 0, 0)

    def unitRegistry(self) -> UnitRegistry:
        """Return the pint UnitRegistry used by this widget."""
        return self._ureg

    def _update_units_combo_choices(self):
        if self._value.dimensionless:
            with signals_blocked(self._units_combo):
                self._units_combo.clear()
                self._units_combo.addItem(NULL_OPTION)
                self._units_combo.addItems(
                    [self._format_units(x) for x in SI_BASES.values()]
                )
                self._units_combo.setCurrentText(NULL_OPTION)
            return

        units = self._value.units
        dims, exp = next(iter(units.dimensionality.items()))
        if exp != 1:
            raise NotImplementedError("Inverse units not yet implemented")
        options = [
            self._format_units(self._ureg.Unit(u))
            for u in DEFAULT_OPTIONS.get(dims, [])
        ]
        current = self._format_units(units)
        with signals_blocked(self._units_combo):
            self._units_combo.clear()
            self._units_combo.addItems(options)
            if self._units_combo.findText(current) == -1:
                self._units_combo.addItem(current)

        self._units_combo.setCurrentText(current)

    def value(self) -> Quantity:
        """Return the current value as a `pint.Quantity`."""
        return self._value

    def text(self) -> str:
        return str(self._value)

    def magnitude(self) -> float | int:
        """Return the magnitude of the current value."""
        return self._value.magnitude

    def units(self) -> Unit:
        """Return the current units."""
        return self._value.units

    def dimensionality(self) -> UnitsContainer:
        """Return the current dimensionality (cast to `str` for nice repr)."""
        return self._value.dimensionality

    def setDecimals(self, decimals: int) -> None:
        """Set the number of decimals to display in the spinbox."""
        self._mag_spinbox.setDecimals(decimals)
        if self._value is not None:
            self._mag_spinbox.setValue(self._value.magnitude)

    def setValue(
        self,
        value: str | Quantity | Number,
        units: UnitsContainer | str | Quantity | None = None,
    ) -> None:
        """Set the current value (will cast to a pint Quantity)."""
        if isinstance(value, Quantity):
            if units is not None:
                raise ValueError("Cannot specify units if value is a Quantity")
            new_val = self._ureg.Quantity(value.magnitude, units=value.units)
        else:
            new_val = self._ureg.Quantity(value, units=units)

        mag_change = new_val.magnitude != self._value.magnitude
        units_change = new_val.units != self._value.units
        dims_changed = new_val.dimensionality != self._value.dimensionality

        self._value = new_val

        if mag_change:
            with signals_blocked(self._mag_spinbox):
                self._mag_spinbox.setValue(float(self._value.magnitude))

        if units_change:
            with signals_blocked(self._units_combo):
                self._units_combo.setCurrentText(self._format_units(self._value.units))
            self.unitsChanged.emit(self._value.units)

        if dims_changed:
            self._update_units_combo_choices()
            self.dimensionalityChanged.emit(self._value.dimensionality)

        if mag_change or units_change:
            self.valueChanged.emit(self._value)

    def setMagnitude(self, magnitude: Number) -> None:
        """Set the magnitude of the current value."""
        self.setValue(self._ureg.Quantity(magnitude, self._value.units))

    def setUnits(self, units: str | Unit | Quantity) -> None:
        """Set the units of the current value.

        If `units` is `None`, will convert to a dimensionless quantity.
        Otherwise, units must be compatible with the current dimensionality.
        """
        if units is None:
            new_val = self._ureg.Quantity(self._value.magnitude)
        elif self.isDimensionless():
            new_val = self._ureg.Quantity(self._value.magnitude, units)
        else:
            new_val = self._value.to(units)
        self.setValue(new_val)

    def isDimensionless(self) -> bool:
        """Return `True` if the current value is dimensionless."""
        return self._value.dimensionless

    def magnitudeSpinBox(self) -> QDoubleSpinBox:
        """Return the `QSpinBox` widget used to edit the magnitude."""
        return self._mag_spinbox

    def unitsComboBox(self) -> QComboBox:
        """Return the `QCombBox` widget used to edit the units."""
        return self._units_combo

    def _format_units(self, u: Unit | str) -> str:
        if isinstance(u, str):
            return u
        return f"{u:~}" if self._abbreviate_units else f"{u:}"



"""A collapsible widget to hide and unhide child widgets."""

from __future__ import annotations

from qtpy.QtCore import (
    QEasingCurve,
    QEvent,
    QMargins,
    QObject,
    QPropertyAnimation,
    QRect,
    QSignalBlocker,
    Qt,
    Signal,
)
from qtpy.QtGui import QIcon, QPainter, QPalette, QPixmap
from qtpy.QtWidgets import QFrame, QPushButton, QSizePolicy, QVBoxLayout, QWidget


class QCollapsible(QFrame):
    """A collapsible widget to hide and unhide child widgets.

    A signal is emitted when the widget is expanded (True) or collapsed (False).

    Based on https://stackoverflow.com/a/68141638
    """

    toggled = Signal(bool)

    def __init__(
        self,
        title: str = "",
        parent: QWidget | None = None,
        expandedIcon: QIcon | str | None = "▼",
        collapsedIcon: QIcon | str | None = "▲",
    ):
        super().__init__(parent)
        self._locked = False
        self._is_animating = False
        self._text = title

        self._toggle_btn = QPushButton(title)
        self._toggle_btn.setCheckable(True)
        self.setCollapsedIcon(icon=collapsedIcon)
        self.setExpandedIcon(icon=expandedIcon)
        self.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Maximum)

        self._toggle_btn.setStyleSheet("text-align: left; border: none; outline: none;")
        self._toggle_btn.toggled.connect(self._toggle)

        # frame layout
        layout = QVBoxLayout(self)
        layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        layout.addWidget(self._toggle_btn)

        # Create animators
        self._animation = QPropertyAnimation(self)
        self._animation.setPropertyName(b"maximumHeight")
        self._animation.setStartValue(0)
        self._animation.finished.connect(self._on_animation_done)
        self.setDuration(300)
        self.setEasingCurve(QEasingCurve.Type.InOutCubic)

        # default content widget
        _content = QWidget()
        _content.setLayout(QVBoxLayout())
        _content.setMaximumHeight(0)
        _content.layout().setContentsMargins(QMargins(5, 0, 0, 0))
        self.setContent(_content)

    def toggleButton(self) -> QPushButton:
        """Return the toggle button."""
        return self._toggle_btn

    def setText(self, text: str) -> None:
        """Set the text of the toggle button."""
        self._toggle_btn.setText(text)

    def text(self) -> str:
        """Return the text of the toggle button."""
        return self._toggle_btn.text()

    def setContent(self, content: QWidget) -> None:
        """Replace central widget (the widget that gets expanded/collapsed)."""
        self._content = content
        self.layout().addWidget(self._content)
        self._animation.setTargetObject(content)

    def content(self) -> QWidget:
        """Return the current content widget."""
        return self._content

    def _convert_string_to_icon(self, symbol: str) -> QIcon:
        """Create a QIcon from a string."""
        size = self._toggle_btn.font().pointSize()
        pixmap = QPixmap(size, size)
        pixmap.fill(Qt.GlobalColor.transparent)
        painter = QPainter(pixmap)
        color = self._toggle_btn.palette().color(QPalette.ColorRole.WindowText)
        painter.setPen(color)
        painter.drawText(QRect(0, 0, size, size), Qt.AlignmentFlag.AlignCenter, symbol)
        painter.end()
        return QIcon(pixmap)

    def expandedIcon(self) -> QIcon:
        """Returns the icon used when the widget is expanded."""
        return self._expanded_icon

    def setExpandedIcon(self, icon: QIcon | str | None = None) -> None:
        """Set the icon on the toggle button when the widget is expanded."""
        if icon and isinstance(icon, QIcon):
            self._expanded_icon = icon
        elif icon and isinstance(icon, str):
            self._expanded_icon = self._convert_string_to_icon(icon)

        if self.isExpanded():
            self._toggle_btn.setIcon(self._expanded_icon)

    def collapsedIcon(self) -> QIcon:
        """Returns the icon used when the widget is collapsed."""
        return self._collapsed_icon

    def setCollapsedIcon(self, icon: QIcon | str | None = None) -> None:
        """Set the icon on the toggle button when the widget is collapsed."""
        if icon and isinstance(icon, QIcon):
            self._collapsed_icon = icon
        elif icon and isinstance(icon, str):
            self._collapsed_icon = self._convert_string_to_icon(icon)

        if not self.isExpanded():
            self._toggle_btn.setIcon(self._collapsed_icon)

    def setDuration(self, msecs: int) -> None:
        """Set duration of the collapse/expand animation."""
        self._animation.setDuration(msecs)

    def setEasingCurve(self, easing: QEasingCurve | QEasingCurve.Type) -> None:
        """Set the easing curve for the collapse/expand animation."""
        self._animation.setEasingCurve(easing)

    def addWidget(self, widget: QWidget) -> None:
        """Add a widget to the central content widget's layout."""
        widget.installEventFilter(self)
        self._content.layout().addWidget(widget)

    def removeWidget(self, widget: QWidget) -> None:
        """Remove widget from the central content widget's layout."""
        self._content.layout().removeWidget(widget)
        widget.removeEventFilter(self)

    def expand(self, animate: bool = True) -> None:
        """Expand (show) the collapsible section."""
        with QSignalBlocker(self._toggle_btn):
            self._expand_collapse(QPropertyAnimation.Direction.Forward, animate)

    def collapse(self, animate: bool = True) -> None:
        """Collapse (hide) the collapsible section."""
        with QSignalBlocker(self._toggle_btn):
            self._expand_collapse(QPropertyAnimation.Direction.Backward, animate)

    def isExpanded(self) -> bool:
        """Return whether the collapsible section is visible."""
        return self._toggle_btn.isChecked()

    def setLocked(self, locked: bool = True) -> None:
        """Set whether collapse/expand is disabled."""
        self._locked = locked
        self._toggle_btn.setCheckable(not locked)

    def locked(self) -> bool:
        """Return True if collapse/expand is disabled."""
        return self._locked

    def _expand_collapse(
        self,
        direction: QPropertyAnimation.Direction,
        animate: bool = True,
        emit: bool = True,
    ) -> None:
        """Set values for the widget based on whether it is expanding or collapsing.

        An emit flag is included so that the toggle signal is only called once (it
        was being emitted a few times via eventFilter when the widget was expanding
        previously).
        """
        if self._locked:
            return

        forward = direction == QPropertyAnimation.Direction.Forward
        icon = self._expanded_icon if forward else self._collapsed_icon
        self._toggle_btn.setIcon(icon)
        self._toggle_btn.setChecked(forward)

        _content_height = self._content.sizeHint().height() + 10
        if animate:
            self._animation.setDirection(direction)
            self._animation.setEndValue(_content_height)
            self._is_animating = True
            self._animation.start()
        else:
            self._content.setMaximumHeight(_content_height if forward else 0)
        if emit:
            self.toggled.emit(direction == QPropertyAnimation.Direction.Forward)

    def _toggle(self) -> None:
        self.expand() if self.isExpanded() else self.collapse()

    def eventFilter(self, a0: QObject, a1: QEvent) -> bool:
        """If a child widget resizes, we need to update our expanded height."""
        if (
            a1.type() == QEvent.Type.Resize
            and self.isExpanded()
            and not self._is_animating
        ):
            self._expand_collapse(
                QPropertyAnimation.Direction.Forward, animate=False, emit=False
            )
        return False

    def _on_animation_done(self) -> None:
        self._is_animating = False



import logging
from collections.abc import Iterable, Mapping
from typing import Any

from qtpy.QtCore import QRegularExpression
from qtpy.QtWidgets import QLineEdit, QTreeWidget, QTreeWidgetItem, QVBoxLayout, QWidget


class QSearchableTreeWidget(QWidget):
    """A tree widget for showing a mapping that can be searched by key.

    This is intended to be used with a read-only mapping and be conveniently
    created using `QSearchableTreeWidget.fromData(data)`.
    If the mapping changes, the easiest way to update this is by calling `setData`.

    The tree can be searched by entering a regular expression pattern
    into the `filter` line edit. An item is only shown if its, any of its ancestors',
    or any of its descendants' keys or values match this pattern.
    The regular expression follows the conventions described by the Qt docs:
    https://doc.qt.io/qt-6/qregularexpression.html#details

    Attributes
    ----------
    tree : QTreeWidget
        Shows the mapping as a tree of items.
    filter : QLineEdit
        Used to filter items in the tree by matching their key against a
        regular expression.
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        self.tree: QTreeWidget = QTreeWidget(self)
        self.tree.setHeaderLabels(("Key", "Value"))

        self.filter: QLineEdit = QLineEdit(self)
        self.filter.setClearButtonEnabled(True)
        self.filter.textChanged.connect(self._updateVisibleItems)

        layout = QVBoxLayout(self)
        layout.addWidget(self.filter)
        layout.addWidget(self.tree)

    def setData(self, data: Mapping) -> None:
        """Update the mapping data shown by the tree."""
        self.tree.clear()
        self.filter.clear()
        top_level_items = [_make_item(name=k, value=v) for k, v in data.items()]
        self.tree.addTopLevelItems(top_level_items)

    def _updateVisibleItems(self, pattern: str) -> None:
        """Recursively update the visibility of items based on the given pattern."""
        expression = QRegularExpression(pattern)
        for i in range(self.tree.topLevelItemCount()):
            top_level_item = self.tree.topLevelItem(i)
            _update_visible_items(top_level_item, expression)

    @classmethod
    def fromData(
        cls, data: Mapping, *, parent: QWidget = None
    ) -> "QSearchableTreeWidget":
        """Make a searchable tree widget from a mapping."""
        widget = cls(parent)
        widget.setData(data)
        return widget


def _make_item(*, name: str, value: Any) -> QTreeWidgetItem:
    """Make a tree item where the name and value are two columns.

    Iterable values other than strings are recursively traversed to
    add child items and build a tree. In this case, mappings use keys
    as their names whereas other iterables use their enumerated index.
    """
    if isinstance(value, Mapping):
        item = QTreeWidgetItem([name, type(value).__name__])
        for k, v in value.items():
            child = _make_item(name=k, value=v)
            item.addChild(child)
    elif isinstance(value, Iterable) and not isinstance(value, str):
        item = QTreeWidgetItem([name, type(value).__name__])
        for i, v in enumerate(value):
            child = _make_item(name=str(i), value=v)
            item.addChild(child)
    else:
        item = QTreeWidgetItem([name, str(value)])
    logging.debug("_make_item: %s, %s, %s", item.text(0), item.text(1), item.flags())
    return item


def _update_visible_items(
    item: QTreeWidgetItem, expression: QRegularExpression, ancestor_match: bool = False
) -> bool:
    """Recursively update the visibility of a tree item based on an expression.

    An item is visible if any of its, any of its ancestors', or any of its descendants'
    column's text matches the expression.
    Returns True if the item is visible, False otherwise.
    """
    match = ancestor_match or any(
        expression.match(item.text(i)).hasMatch() for i in range(item.columnCount())
    )
    visible = match
    for i in range(item.childCount()):
        child = item.child(i)
        descendant_visible = _update_visible_items(child, expression, match)
        visible = visible or descendant_visible
    item.setHidden(not visible)
    logging.debug(
        "_update_visible_items: %s, %s",
        tuple(item.text(i) for i in range(item.columnCount())),
        visible,
    )
    return visible



"""superqt is a collection of Qt components for python."""

from importlib.metadata import PackageNotFoundError, version
from typing import TYPE_CHECKING, Any

try:
    __version__ = version("superqt")
except PackageNotFoundError:
    __version__ = "unknown"

from .collapsible import QCollapsible
from .combobox import QColorComboBox, QEnumComboBox, QSearchableComboBox
from .elidable import QElidingLabel, QElidingLineEdit
from .selection import QSearchableListWidget, QSearchableTreeWidget
from .sliders import (
    QDoubleRangeSlider,
    QDoubleSlider,
    QLabeledDoubleRangeSlider,
    QLabeledDoubleSlider,
    QLabeledRangeSlider,
    QLabeledSlider,
    QRangeSlider,
)
from .spinbox import QLargeIntSpinBox
from .switch import QToggleSwitch
from .utils import (
    QFlowLayout,
    QMessageHandler,
    ensure_main_thread,
    ensure_object_thread,
)

__all__ = [
    "QCollapsible",
    "QColorComboBox",
    "QColormapComboBox",
    "QDoubleRangeSlider",
    "QDoubleSlider",
    "QElidingLabel",
    "QElidingLineEdit",
    "QEnumComboBox",
    "QFlowLayout",
    "QIconifyIcon",
    "QLabeledDoubleRangeSlider",
    "QLabeledDoubleSlider",
    "QLabeledRangeSlider",
    "QLabeledSlider",
    "QLargeIntSpinBox",
    "QMessageHandler",
    "QQuantity",
    "QRangeSlider",
    "QSearchableComboBox",
    "QSearchableListWidget",
    "QSearchableTreeWidget",
    "QToggleSwitch",
    "ensure_main_thread",
    "ensure_object_thread",
]

if TYPE_CHECKING:
    from .combobox import QColormapComboBox
    from .iconify import QIconifyIcon
    from .spinbox._quantity import QQuantity


def __getattr__(name: str) -> Any:
    if name == "QColormapComboBox":
        from .cmap import QColormapComboBox

        return QColormapComboBox
    if name == "QIconifyIcon":
        from .iconify import QIconifyIcon

        return QIconifyIcon
    if name == "QQuantity":
        from .spinbox._quantity import QQuantity

        return QQuantity
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



from __future__ import annotations

from enum import IntEnum, IntFlag, auto
from typing import TYPE_CHECKING, Any, overload

from qtpy import QtGui
from qtpy.QtCore import Property, QPoint, QSize, Qt, Signal
from qtpy.QtGui import QDoubleValidator, QFontMetrics, QValidator
from qtpy.QtWidgets import (
    QAbstractSlider,
    QBoxLayout,
    QHBoxLayout,
    QLineEdit,
    QSlider,
    QStyle,
    QStyleOption,
    QVBoxLayout,
    QWidget,
)

from superqt.utils import signals_blocked

from ._sliders import QDoubleRangeSlider, QDoubleSlider, QRangeSlider

if TYPE_CHECKING:
    from collections.abc import Iterable


class LabelPosition(IntEnum):
    NoLabel = 0
    LabelsAbove = auto()
    LabelsBelow = auto()
    LabelsRight = LabelsAbove
    LabelsLeft = LabelsBelow
    LabelsOnHandle = auto()


class EdgeLabelMode(IntFlag):
    NoLabel = 0
    LabelIsRange = auto()
    LabelIsValue = auto()


class _SliderProxy:
    _slider: QSlider

    def value(self) -> Any:
        return self._slider.value()

    def setValue(self, value: Any) -> None:
        self._slider.setValue(value)

    def sliderPosition(self) -> int:
        return self._slider.sliderPosition()

    def setSliderPosition(self, pos: int) -> None:
        self._slider.setSliderPosition(pos)

    def minimum(self) -> int:
        return self._slider.minimum()

    def setMinimum(self, minimum: int) -> None:
        self._slider.setMinimum(minimum)

    def maximum(self) -> int:
        return self._slider.maximum()

    def setMaximum(self, maximum: int) -> None:
        self._slider.setMaximum(maximum)

    def singleStep(self):
        return self._slider.singleStep()

    def setSingleStep(self, step: int) -> None:
        self._slider.setSingleStep(step)

    def pageStep(self) -> int:
        return self._slider.pageStep()

    def setPageStep(self, step: int) -> None:
        self._slider.setPageStep(step)

    def setRange(self, min: float, max: float) -> None:
        self._slider.setRange(min, max)

    def tickInterval(self) -> int:
        return self._slider.tickInterval()

    def setTickInterval(self, interval: int) -> None:
        self._slider.setTickInterval(interval)

    def tickPosition(self) -> QSlider.TickPosition:
        return self._slider.tickPosition()

    def setTickPosition(self, pos: QSlider.TickPosition) -> None:
        self._slider.setTickPosition(pos)

    def triggerAction(self, action: QAbstractSlider.SliderAction) -> None:
        return self._slider.triggerAction(action)

    def invertedControls(self) -> bool:
        return self._slider.invertedControls()

    def setInvertedControls(self, a0: bool) -> None:
        return self._slider.setInvertedControls(a0)

    def invertedAppearance(self) -> bool:
        return self._slider.invertedAppearance()

    def setInvertedAppearance(self, a0: bool) -> None:
        return self._slider.setInvertedAppearance(a0)

    def isSliderDown(self) -> bool:
        return self._slider.isSliderDown()

    def setSliderDown(self, a0: bool) -> None:
        return self._slider.setSliderDown(a0)

    def hasTracking(self) -> bool:
        return self._slider.hasTracking()

    def setTracking(self, enable: bool) -> None:
        return self._slider.setTracking(enable)

    def orientation(self) -> Qt.Orientation:
        return self._slider.orientation()

    def __getattr__(self, name: Any) -> Any:
        return getattr(self._slider, name)


def _handle_overloaded_slider_sig(
    args: tuple, kwargs: dict
) -> tuple[QWidget | None, Qt.Orientation]:
    """Maintaining signature of QSlider.__init__."""
    parent = None
    orientation = Qt.Orientation.Horizontal
    errmsg = (
        "TypeError: arguments did not match any overloaded call:\n"
        "  QSlider(parent: QWidget = None)\n"
        "  QSlider(Qt.Orientation, parent: QWidget = None)"
    )
    if len(args) > 2:
        raise TypeError(errmsg)
    elif len(args) == 2:
        if kwargs:
            raise TypeError(errmsg)
        orientation, parent = args
    elif args:
        if isinstance(args[0], QWidget):
            if kwargs:
                raise TypeError(errmsg)
            parent = args[0]
        else:
            orientation = args[0]
    parent = kwargs.get("parent", parent)
    return parent, orientation


class QLabeledSlider(_SliderProxy, QAbstractSlider):
    editingFinished = Signal()

    _slider_class = QSlider
    _slider: QSlider

    @overload
    def __init__(self, parent: QWidget | None = ...) -> None: ...

    @overload
    def __init__(
        self, orientation: Qt.Orientation, parent: QWidget | None = ...
    ) -> None: ...

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        parent, orientation = _handle_overloaded_slider_sig(args, kwargs)

        super().__init__(parent)
        # accept focus events
        fp = self.style().styleHint(QStyle.StyleHint.SH_Button_FocusPolicy)
        self.setFocusPolicy(Qt.FocusPolicy(fp))

        self._slider = self._slider_class(parent=self)
        self._label = SliderLabel(self._slider, parent=self)
        self._label.valueEdited.connect(self._setValue)
        self._edge_label_mode: EdgeLabelMode = EdgeLabelMode.LabelIsValue
        self._edge_label_position: LabelPosition = LabelPosition.LabelsRight

        self._rename_signals()
        self._slider.actionTriggered.connect(self.actionTriggered.emit)
        self._slider.rangeChanged.connect(self._on_slider_range_changed)
        self._slider.sliderMoved.connect(self.sliderMoved.emit)
        self._slider.sliderPressed.connect(self.sliderPressed.emit)
        self._slider.sliderReleased.connect(self.sliderReleased.emit)
        self._slider.valueChanged.connect(self._on_slider_value_changed)
        self._label.editingFinished.connect(self.editingFinished)

        self.setOrientation(orientation)

    # ------------------- public API -------------------

    def setOrientation(self, orientation: Qt.Orientation) -> None:
        """Set orientation, value will be 'horizontal' or 'vertical'."""
        self._slider.setOrientation(orientation)
        marg = (0, 0, 0, 0)
        if orientation == Qt.Orientation.Vertical:
            layout = QVBoxLayout()
            if not self._edge_label_position:
                layout.addWidget(self._slider, alignment=Qt.AlignmentFlag.AlignHCenter)
            elif self._edge_label_position == LabelPosition.LabelsBelow:
                layout.addWidget(self._slider, alignment=Qt.AlignmentFlag.AlignHCenter)
                layout.addWidget(self._label, alignment=Qt.AlignmentFlag.AlignHCenter)
            else:
                layout.addWidget(self._label, alignment=Qt.AlignmentFlag.AlignHCenter)
                layout.addWidget(self._slider, alignment=Qt.AlignmentFlag.AlignHCenter)
            self._label.setAlignment(Qt.AlignmentFlag.AlignCenter)
            layout.setSpacing(1)
        else:
            layout = QHBoxLayout()  # type: ignore
            if not self._edge_label_position:
                layout.addWidget(self._slider)
            elif self._edge_label_position == LabelPosition.LabelsRight:
                layout.addWidget(self._slider)
                layout.addWidget(self._label)
                self._label.setAlignment(Qt.AlignmentFlag.AlignRight)
            else:
                layout.addWidget(self._label)
                layout.addWidget(self._slider)
                self._label.setAlignment(Qt.AlignmentFlag.AlignLeft)
                marg = (0, 0, 5, 0)
            layout.setSpacing(6)

        old_layout = self.layout()
        if old_layout is not None:
            QWidget().setLayout(old_layout)

        layout.setContentsMargins(*marg)
        self.setLayout(layout)

    def edgeLabelMode(self) -> EdgeLabelMode:
        """Return current `EdgeLabelMode`."""
        return self._edge_label_mode

    def setEdgeLabelMode(self, opt: EdgeLabelMode) -> None:
        """Set the `EdgeLabelMode`.

        Parameters
        ----------
        opt : EdgeLabelMode
            To show no label, use `EdgeLabelMode.NoLabel`. To show the value
            of the slider, use `EdgeLabelMode.LabelIsValue`. To show
            `value / maximum`, use
            `EdgeLabelMode.LabelIsValue | EdgeLabelMode.LabelIsRange`.
        """
        if opt is EdgeLabelMode.LabelIsRange:
            raise ValueError(
                "mode must be one of 'EdgeLabelMode.NoLabel' or "
                "'EdgeLabelMode.LabelIsValue' or"
                "'EdgeLabelMode.LabelIsValue | EdgeLabelMode.LabelIsRange'."
            )

        self._edge_label_mode = opt
        self._on_slider_range_changed(self.minimum(), self.maximum())
        if not self._edge_label_mode:
            self._label.hide()
            w = 5 if self.orientation() == Qt.Orientation.Horizontal else 0
            if self._edge_label_position == LabelPosition.LabelsRight:
                self.layout().setContentsMargins(0, 0, w, 0)
            elif self._edge_label_position == LabelPosition.LabelsLeft:
                self.layout().setContentsMargins(0, 0, 0, w)
        if opt & EdgeLabelMode.LabelIsValue:
            if self.isVisible():
                self._label.show()
            self._label.setMode(opt)
            self._label.setRange(self._slider.minimum(), self._slider.maximum())
            self._label.setValue(self._slider.value())
            self.layout().setContentsMargins(0, 0, 0, 0)

    def edgeLabelPosition(self) -> LabelPosition:
        """Return where/whether a label is shown at the edge of the slider."""
        return self._edge_label_position

    def setEdgeLabelPosition(self, opt: LabelPosition) -> None:
        """Set where/whether a label is shown at the edge of the slider."""
        if opt is LabelPosition.LabelsOnHandle:
            raise ValueError("position cannot be 'LabelPosition.LabelsOnHandle'")

        self._edge_label_position = opt
        self._label.setVisible(bool(opt))
        # TODO: make double clickable to edit
        self.setOrientation(self.orientation())

    # putting this after labelMode methods for the sake of mypy
    EdgeLabelMode = EdgeLabelMode
    LabelPosition = LabelPosition

    # --------------------- private api --------------------

    def _on_slider_range_changed(self, min_: int, max_: int) -> None:
        if self._edge_label_mode & EdgeLabelMode.LabelIsRange:
            self._label.setSuffix(f" / {max_}")
        else:
            self._label.setSuffix("")
        self.rangeChanged.emit(min_, max_)

    def _on_slider_value_changed(self, v: Any) -> None:
        self._label.setValue(v)
        self.valueChanged.emit(v)

    def _setValue(self, value: float) -> None:
        """Convert the value from float to int before setting the slider value."""
        self._slider.setValue(int(value))

    def _rename_signals(self) -> None: ...


class QLabeledDoubleSlider(QLabeledSlider):
    _slider_class = QDoubleSlider
    _slider: QDoubleSlider
    fvalueChanged = Signal(float)
    fsliderMoved = Signal(float)
    frangeChanged = Signal(float, float)

    @overload
    def __init__(self, parent: QWidget | None = ...) -> None: ...

    @overload
    def __init__(
        self, orientation: Qt.Orientation, parent: QWidget | None = ...
    ) -> None: ...

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.setDecimals(2)

    def _setValue(self, value: float) -> None:
        """Convert the value from float to int before setting the slider value."""
        self._slider.setValue(value)

    def _rename_signals(self) -> None:
        self.valueChanged = self.fvalueChanged
        self.sliderMoved = self.fsliderMoved
        self.rangeChanged = self.frangeChanged

    def decimals(self) -> int:
        return self._label.decimals()

    def setDecimals(self, prec: int) -> None:
        self._label.setDecimals(prec)


class QLabeledRangeSlider(_SliderProxy, QAbstractSlider):
    valuesChanged = Signal(tuple)
    editingFinished = Signal()

    _slider_class = QRangeSlider
    _slider: QRangeSlider

    @overload
    def __init__(self, parent: QWidget | None = ...) -> None: ...

    @overload
    def __init__(
        self, orientation: Qt.Orientation, parent: QWidget | None = ...
    ) -> None: ...

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        parent, orientation = _handle_overloaded_slider_sig(args, kwargs)
        super().__init__(parent)
        self._rename_signals()

        self.setAttribute(Qt.WidgetAttribute.WA_ShowWithoutActivating)
        self._handle_labels: list[SliderLabel] = []
        self._handle_label_position: LabelPosition = LabelPosition.LabelsAbove

        # for fine tuning label position
        self.label_shift_x = 0
        self.label_shift_y = 0

        self._slider = self._slider_class()
        self._slider.valueChanged.connect(self.valueChanged.emit)
        self._slider.sliderPressed.connect(self.sliderPressed.emit)
        self._slider.sliderReleased.connect(self.sliderReleased.emit)
        self._slider.rangeChanged.connect(self.rangeChanged.emit)
        self.sliderMoved = self._slider.slidersMoved

        self._min_label = SliderLabel(
            self._slider, alignment=Qt.AlignmentFlag.AlignLeft
        )
        self._min_label.valueEdited.connect(self._min_label_edited)
        self._max_label = SliderLabel(
            self._slider, alignment=Qt.AlignmentFlag.AlignRight
        )
        self._max_label.valueEdited.connect(self._max_label_edited)
        self._min_label.editingFinished.connect(self.editingFinished)
        self._max_label.editingFinished.connect(self.editingFinished)
        self.setEdgeLabelMode(EdgeLabelMode.LabelIsRange)

        self._slider.valueChanged.connect(self._on_value_changed)
        self._slider.rangeChanged.connect(self._on_range_changed)

        self._on_value_changed(self._slider.value())
        self._on_range_changed(self._slider.minimum(), self._slider.maximum())
        self.setOrientation(orientation)

    # --------------------- public API -------------------

    def handleLabelPosition(self) -> LabelPosition:
        """Return where/whether labels are shown adjacent to slider handles."""
        return self._handle_label_position

    def setHandleLabelPosition(self, opt: LabelPosition) -> None:
        """Set where/whether labels are shown adjacent to slider handles."""
        self._handle_label_position = opt
        for lbl in self._handle_labels:
            lbl.setVisible(bool(opt))
            trans = opt == LabelPosition.LabelsOnHandle
            # TODO: make double clickable to edit
            lbl.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents, trans)
        self.setOrientation(self.orientation())

    def edgeLabelMode(self) -> EdgeLabelMode:
        """Return current `EdgeLabelMode`."""
        return self._edge_label_mode

    def setEdgeLabelMode(self, opt: EdgeLabelMode) -> None:
        """Set `EdgeLabelMode`, controls what is shown at the min/max labels."""
        self._edge_label_mode = opt
        if not self._edge_label_mode:
            self._min_label.hide()
            self._max_label.hide()
        else:
            if self.isVisible():
                self._min_label.show()
                self._max_label.show()
            self._min_label.setMode(opt)
            self._min_label.setRange(self._slider.minimum(), self._slider.maximum())
            self._max_label.setMode(opt)
            self._max_label.setRange(self._slider.minimum(), self._slider.maximum())
        if opt == EdgeLabelMode.LabelIsValue:
            v0, *_, v1 = self._slider.value()
            self._min_label.setValue(v0)
            self._max_label.setValue(v1)
        elif opt == EdgeLabelMode.LabelIsRange:
            self._min_label.setValue(self._slider.minimum())
            self._max_label.setValue(self._slider.maximum())
        self._reposition_labels()

    def setRange(self, min: int, max: int) -> None:
        self._on_range_changed(min, max)

    def _add_labels(self, layout: QBoxLayout, inverted: bool = False) -> None:
        if inverted:
            first, second = self._max_label, self._min_label
        else:
            first, second = self._min_label, self._max_label
        layout.addWidget(first)
        layout.addWidget(self._slider)
        layout.addWidget(second)

    def setOrientation(self, orientation: Qt.Orientation) -> None:
        """Set orientation, value will be 'horizontal' or 'vertical'."""
        self._slider.setOrientation(orientation)
        inverted = self._slider.invertedAppearance()
        marg = (0, 0, 0, 0)
        if orientation == Qt.Orientation.Vertical:
            layout: QBoxLayout = QVBoxLayout()
            layout.setSpacing(1)
            self._add_labels(layout, inverted=not inverted)
            # TODO: set margins based on label width
            if self._handle_label_position == LabelPosition.LabelsLeft:
                marg = (30, 0, 0, 0)
            elif self._handle_label_position == LabelPosition.LabelsRight:
                marg = (0, 0, 20, 0)
            layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        else:
            layout = QHBoxLayout()
            layout.setSpacing(7)
            if self._handle_label_position == LabelPosition.LabelsBelow:
                marg = (0, 0, 0, 25)
            elif self._handle_label_position == LabelPosition.LabelsAbove:
                marg = (0, 25, 0, 0)
            self._add_labels(layout, inverted=inverted)

        # remove old layout
        old_layout = self.layout()
        if old_layout is not None:
            QWidget().setLayout(old_layout)

        self.setLayout(layout)
        layout.setContentsMargins(*marg)
        super().setOrientation(orientation)
        self._reposition_labels()

    def setInvertedAppearance(self, a0: bool) -> None:
        self._slider.setInvertedAppearance(a0)
        self.setOrientation(self._slider.orientation())

    def resizeEvent(self, a0: Any) -> None:
        super().resizeEvent(a0)
        self._reposition_labels()

    # putting this after methods above for the sake of mypy
    LabelPosition = LabelPosition
    EdgeLabelMode = EdgeLabelMode

    def _getBarColor(self) -> QtGui.QBrush:
        return self._slider._style.brush(self._slider._styleOption)

    def _setBarColor(self, color: str) -> None:
        self._slider._style.brush_active = color

    barColor = Property(QtGui.QBrush, _getBarColor, _setBarColor)
    """The color of the bar between the first and last handle."""

    # ------------- private methods ----------------
    def _rename_signals(self) -> None:
        self.valueChanged = self.valuesChanged

    def _reposition_labels(self) -> None:
        if (
            not self._handle_labels
            or self._handle_label_position == LabelPosition.NoLabel
        ):
            return

        horizontal = self.orientation() == Qt.Orientation.Horizontal
        labels_above = self._handle_label_position == LabelPosition.LabelsAbove
        labels_on_handle = self._handle_label_position == LabelPosition.LabelsOnHandle

        last_edge = None
        labels: Iterable[tuple[int, SliderLabel]] = enumerate(self._handle_labels)
        if self._slider.invertedAppearance():
            labels = reversed(list(labels))
        for i, label in labels:
            rect = self._slider._handleRect(i)
            dx = (-label.width() / 2) + 2
            dy = -label.height() / 2
            if labels_above:  # or on the right
                if horizontal:
                    dy *= 3
                else:
                    dx *= -1
            elif labels_on_handle:
                if horizontal:
                    dy += 0.5
                else:
                    dx += 0.5
            else:
                if horizontal:
                    dy *= -1
                else:
                    dx *= 3
            pos = self._slider.mapToParent(rect.center())
            pos += QPoint(int(dx + self.label_shift_x), int(dy + self.label_shift_y))
            if last_edge is not None:
                # prevent label overlap
                if horizontal:
                    pos.setX(int(max(pos.x(), last_edge.x() + label.width() / 2 + 12)))
                else:
                    pos.setY(int(min(pos.y(), last_edge.y() - label.height() / 2 - 4)))
            label.move(pos)
            last_edge = pos
            label.clearFocus()
            label.raise_()
            label.show()
        self.update()

    def _min_label_edited(self, val: float) -> None:
        if self._edge_label_mode == EdgeLabelMode.LabelIsRange:
            self.setMinimum(val)
        else:
            v = list(self._slider.value())
            v[0] = val
            self.setValue(v)
        self._reposition_labels()

    def _max_label_edited(self, val: float) -> None:
        if self._edge_label_mode == EdgeLabelMode.LabelIsRange:
            self.setMaximum(val)
        else:
            v = list(self._slider.value())
            v[-1] = val
            self.setValue(v)
        self._reposition_labels()

    def _on_value_changed(self, v: tuple[int, ...]) -> None:
        if self._edge_label_mode == EdgeLabelMode.LabelIsValue:
            self._min_label.setValue(v[0])
            self._max_label.setValue(v[-1])

        if len(v) != len(self._handle_labels):
            for lbl in self._handle_labels:
                lbl.setParent(None)
                lbl.deleteLater()
            self._handle_labels.clear()
            for n, val in enumerate(self._slider.value()):
                s = SliderLabel(self._slider, parent=self, index=n)
                s.valueEdited.connect(self._on_slider_label_edited)
                s.editingFinished.connect(self.editingFinished)
                s.setValue(val)
                self._handle_labels.append(s)
        else:
            for val, label in zip(v, self._handle_labels, strict=False):
                label.setValue(val)
        self._reposition_labels()

    def _on_slider_label_edited(self, pos: float) -> None:
        idx = getattr(self.sender(), "_index", 0)
        self._slider.setSliderPosition(pos, idx)

    def _on_range_changed(self, min: int, max: int) -> None:
        if (min, max) != (self._slider.minimum(), self._slider.maximum()):
            self._slider.setRange(min, max)
        for lbl in self._handle_labels:
            lbl.setRange(min, max)
        if self._edge_label_mode == EdgeLabelMode.LabelIsRange:
            self._min_label.setValue(min)
            self._max_label.setValue(max)
        self._reposition_labels()

    # def setValue(self, value) -> None:
    #     super().setValue(value)
    #     self.sliderChange(QSlider.SliderValueChange)


class QLabeledDoubleRangeSlider(QLabeledRangeSlider):
    _slider_class = QDoubleRangeSlider
    _slider: QDoubleRangeSlider
    frangeChanged = Signal(float, float)

    @overload
    def __init__(self, parent: QWidget | None = ...) -> None: ...

    @overload
    def __init__(
        self, orientation: Qt.Orientation, parent: QWidget | None = ...
    ) -> None: ...

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.setDecimals(2)

    def _rename_signals(self) -> None:
        super()._rename_signals()
        self.rangeChanged = self.frangeChanged

    def decimals(self) -> int:
        return self._min_label.decimals()

    def setDecimals(self, prec: int) -> None:
        self._min_label.setDecimals(prec)
        self._max_label.setDecimals(prec)
        for lbl in self._handle_labels:
            lbl.setDecimals(prec)

    def _getBarColor(self) -> QtGui.QBrush:
        return self._slider._style.brush(self._slider._styleOption)

    def _setBarColor(self, color: str) -> None:
        self._slider._style.brush_active = color

    barColor = Property(QtGui.QBrush, _getBarColor, _setBarColor)
    """The color of the bar between the first and last handle."""


class SliderLabel(QLineEdit):
    valueEdited = Signal(float)

    def __init__(
        self,
        slider: QSlider,
        parent: QWidget | None = None,
        alignment: Qt.AlignmentFlag = Qt.AlignmentFlag.AlignCenter,
        index: int = 0,
    ) -> None:
        super().__init__(parent=parent)
        self._index = index
        self._prefix = ""
        self._suffix = ""
        self._min = slider.minimum()
        self._max = slider.maximum()
        self._value = self._min
        self._decimals = -1
        self.setFocusPolicy(Qt.FocusPolicy.ClickFocus)
        self.setMode(EdgeLabelMode.LabelIsValue)
        self.setDecimals(0)  # calls updateText
        validator = QDoubleValidator(self)
        validator.setNotation(QDoubleValidator.Notation.ScientificNotation)
        self.setValidator(validator)

        slider.rangeChanged.connect(self.setRange)
        slider.rangeChanged.connect(self._update_size)
        self.setAlignment(alignment)
        self.setStyleSheet("background:transparent; border: 0;")
        self.editingFinished.connect(self._editing_finished)
        self.editingFinished.connect(self._silent_clear_focus)
        self._update_size()

    def _editing_finished(self):
        self._silent_clear_focus()
        self.setValue(float(self.text()))
        self.valueEdited.emit(self.value())

    def setRange(self, min_: float, max_: float) -> None:
        if self._mode == EdgeLabelMode.LabelIsRange:
            max_val = max(abs(min_), abs(max_))
            n_digits = max(len(str(int(max_val))), 7)
            upper_bound = int("9" * n_digits)
            self._min = -upper_bound
            self._max = upper_bound
            self._update_size()
        else:
            max_ = max(max_, min_)
            self._min = min_
            self._max = max_

    def setDecimals(self, prec: int) -> None:
        # super().setDecimals(prec)
        self._decimals = prec
        self.updateText()
        self._update_size()

    def decimals(self) -> int:
        """Return the number of decimals used in the label."""
        return self._decimals

    def value(self) -> float:
        return self._value

    def setValue(self, val: Any) -> None:
        if val < self._min:
            val = self._min
        elif val > self._max:
            val = self._max
        self._value = val
        self.updateText()

    def updateText(self) -> None:
        val = float(self._value)
        use_scientific = (abs(val) < 0.0001 or abs(val) > 9999999.0) and val != 0.0
        font_metrics = QFontMetrics(self.font())
        eight_len = _fm_width(font_metrics, "8")

        available_chars = self.width() // eight_len

        total, _fraction = f"{val:.<f}".split(".")

        if len(total) > available_chars:
            use_scientific = True

        if self._decimals < 0:
            if use_scientific:
                mantissa, exponent = f"{val:.{available_chars}e}".split("e")
                mantissa = mantissa.rstrip("0").rstrip(".")
                if len(mantissa) + len(exponent) + 1 < available_chars:
                    text = f"{mantissa}e{exponent}"
                else:
                    decimals = max(available_chars - len(exponent) - 3, 2)
                    text = f"{val:.{decimals}e}"

            else:
                decimals = max(available_chars - len(total) - 1, 2)
                text = f"{val:.{decimals}f}"
                text = text.rstrip("0").rstrip(".")
        else:
            if use_scientific:
                mantissa, exponent = f"{val:.{self._decimals}e}".split("e")
                mantissa = mantissa.rstrip("0").rstrip(".")
                text = f"{mantissa}e{exponent}"
            else:
                text = f"{val:.{self._decimals}f}"
        if text == "":
            text = "0"
        self.setText(text)
        if self._mode == EdgeLabelMode.LabelIsRange:
            self._update_size()

    def minimum(self):
        return self._min

    def setMaximum(self, max_: float) -> None:
        self.setRange(self._min, max_)

    def maximum(self):
        return self._max

    def setMinimum(self, min_: float) -> None:
        self.setRange(min_, self._max)

    def setMode(self, opt: EdgeLabelMode) -> None:
        self._mode = opt
        self._update_size()

    def prefix(self) -> str:
        return self._prefix

    def setPrefix(self, prefix: str) -> None:
        self._prefix = prefix
        self._update_size()

    def suffix(self) -> str:
        return self._suffix

    def setSuffix(self, suffix: str) -> None:
        self._suffix = suffix
        self._update_size()

    # --------------- private ----------------

    def _silent_clear_focus(self) -> None:
        with signals_blocked(self):
            self.clearFocus()

    def _get_size(self):
        # fontmetrics to measure the width of text
        fm = QFontMetrics(self.font())
        h = self.sizeHint().height()
        fixed_content = self.prefix() + self.suffix() + " "

        if self._mode & EdgeLabelMode.LabelIsValue:
            # determine width based on min/max/specialValue
            mintext = str(self.minimum())[:18]
            maxtext = str(self.maximum())[:18]
            w = max(0, _fm_width(fm, mintext + fixed_content))
            w = max(w, _fm_width(fm, maxtext + fixed_content))
            if self._mode & EdgeLabelMode.LabelIsRange:
                w += 8  # it seems as thought suffix() is not enough
        else:
            w = max(0, _fm_width(fm, str(self.value()))) + 3

        w += 3  # cursor blinking space
        # get the final size hint
        opt = QStyleOption()
        # self.initStyleOption(opt)
        return self.style().sizeFromContents(
            QStyle.ContentsType.CT_LineEdit, opt, QSize(w, h), self
        )

    def _update_size(self, *_: Any) -> None:
        self.setFixedSize(self._get_size())

    def validate(
        self, input_: str | None, pos: int
    ) -> tuple[QValidator.State, str, int]:
        # fake like an integer spinbox
        if input_ and "." in input_ and self.decimals() < 1:
            return QValidator.State.Invalid, input_, len(input_)
        return super().validate(input_, pos)

    def showEvent(self, event: Any) -> None:
        # need to update size after showing
        # to handle the qss font size change
        super().showEvent(event)
        self._update_size()


def _fm_width(fm: QFontMetrics, text: str) -> int:
    if hasattr(fm, "horizontalAdvance"):
        return fm.horizontalAdvance(text)
    return fm.width(text)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from qtpy.QtCore import QRect, Qt
from qtpy.QtGui import QIcon, QPainter, QPaintEvent, QPalette
from qtpy.QtWidgets import (
    QApplication,
    QLineEdit,
    QStyle,
    QStyleOptionFrame,
    QWidget,
)

from ._cmap_utils import draw_colormap, pick_font_color, try_cast_colormap

if TYPE_CHECKING:
    from cmap import Colormap

MISSING = QStyle.StandardPixmap.SP_TitleBarContextHelpButton


class QColormapLineEdit(QLineEdit):
    """A QLineEdit that shows a colormap swatch.

    When the current text is a valid colormap name from the `cmap` package, a swatch
    of the colormap will be shown to the left of the text (if `fractionalColormapWidth`
    is less than .75) or behind the text (for when the colormap fills the full width).

    If the current text is not a valid colormap name, a swatch of the fallback colormap
    will be shown instead (by default, a gray colormap) if `fractionalColormapWidth` is
    less than .75.

    Parameters
    ----------
    parent : QWidget, optional
        The parent widget.
    fractional_colormap_width : float, optional
        The fraction of the widget width to use for the colormap swatch. If the
        colormap is full width (greater than 0.75), the swatch will be drawn behind
        the text. Otherwise, the swatch will be drawn to the left of the text.
        Default is 0.33.
    fallback_cmap : Colormap | str | None, optional
        The colormap to use when the current text is not a recognized colormap.
        by default "gray".
    missing_icon : QIcon | QStyle.StandardPixmap, optional
        The icon to show when the current text is not a recognized colormap and
        `fractionalColormapWidth` is less than .75. Default is a question mark.
    checkerboard_size : int, optional
        Size (in pixels) of the checkerboard pattern to draw behind colormaps with
        transparency, by default 4. If 0, no checkerboard is drawn.
    allow_invalid : bool, optional
        If True, the user can enter any text, even if it does not represent a valid
        colormap (and `fallback_cmap` will be shown if it's invalid). If False, the text
        will be validated when editing is finished or focus is lost, and if the text is
        not a valid colormap, it will be reverted to the first available valid option
        from the completer, or, if that's not available, the last valid colormap.
        Default is True.  This is only settable at initialization.
    """

    def __init__(
        self,
        parent: QWidget | None = None,
        *,
        fractional_colormap_width: float = 0.35,
        fallback_cmap: Colormap | str | None = "gray",
        missing_icon: QIcon | QStyle.StandardPixmap = MISSING,
        checkerboard_size: int = 4,
        allow_invalid: bool = True,
    ) -> None:
        super().__init__(parent)
        self.setFractionalColormapWidth(fractional_colormap_width)
        self.setMissingColormap(fallback_cmap)
        self._checkerboard_size = checkerboard_size

        if isinstance(missing_icon, QStyle.StandardPixmap):
            self._missing_icon: QIcon = self.style().standardIcon(missing_icon)
        elif isinstance(missing_icon, QIcon):
            self._missing_icon = missing_icon
        else:  # pragma: no cover
            raise TypeError("missing_icon must be a QIcon or QStyle.StandardPixmap")

        self._cmap: Colormap | None = None  # current colormap
        self.textChanged.connect(self.setColormap)

        self._lastValidColormap: Colormap | None = None
        if not allow_invalid:
            self.editingFinished.connect(self._validate)

    def _validate(self) -> None:
        """Called when editing is finished or focus is lost.

        If the current text does not represent a valid colormap, revert to the first
        available valid option from the completer, or, if that's not available, revert
        to the last valid colormap.
        """
        if self._cmap is None:
            candidate = self._fist_completer_option()
            if candidate is not None:
                self.setColormap(candidate)
                self.setText(candidate.name.rsplit(":", 1)[-1])
            elif self._lastValidColormap is not None:
                self.setColormap(self._lastValidColormap)
                self.setText(self._lastValidColormap.name.rsplit(":", 1)[-1])
            # Optionally, if neither is available, you might decide to clear the text.
        else:
            # Update the last valid value.
            self._lastValidColormap = self._cmap

    def _fist_completer_option(self) -> Colormap | None:
        """Return the first valid Colormap from the completer's current filtered list.

        or None if no valid option is available.
        """
        if (
            (completer := self.completer()) is None
            or (model := completer.model()) is None
            or model.rowCount() == 0
        ):
            return None

        first_item = model.index(0, 0).data(Qt.ItemDataRole.DisplayRole)
        return try_cast_colormap(first_item)

    def setFractionalColormapWidth(self, fraction: float) -> None:
        self._colormap_fraction: float = float(fraction)
        align = Qt.AlignmentFlag.AlignVCenter
        if self._cmap_is_full_width():
            align |= Qt.AlignmentFlag.AlignCenter
        else:
            align |= Qt.AlignmentFlag.AlignLeft
        self.setAlignment(align)

    def fractionalColormapWidth(self) -> float:
        return self._colormap_fraction

    def setMissingColormap(self, cmap: Colormap | str | None) -> None:
        self._missing_cmap: Colormap | None = try_cast_colormap(cmap)

    def colormap(self) -> Colormap | None:
        return self._cmap

    def setColormap(self, cmap: Colormap | str | None) -> None:
        self._cmap = try_cast_colormap(cmap)

        # set self font color to contrast with the colormap
        if self._cmap and self._cmap_is_full_width():
            text = pick_font_color(self._cmap)
        else:
            text = QApplication.palette().color(QPalette.ColorRole.Text)

        palette = self.palette()
        palette.setColor(QPalette.ColorRole.Text, text)
        self.setPalette(palette)

    def _cmap_is_full_width(self):
        return self._colormap_fraction >= 0.75

    def _content_rect(self) -> QRect:
        """Return the style-aware content rect for this line edit."""
        opt = QStyleOptionFrame()
        self.initStyleOption(opt)
        return self.style().subElementRect(
            QStyle.SubElement.SE_LineEditContents, opt, self
        )

    def _cmap_rect(self) -> QRect:
        cr = self._content_rect()
        # Apply the horizontal content inset as vertical padding too,
        # so the swatch sits inside the style's painted background.
        v_pad = cr.x() - self.rect().x()
        return QRect(
            max(cr.x(), 2),
            v_pad,
            int(cr.width() * self._colormap_fraction),
            self.rect().height() - 2 * v_pad,
        )

    def resizeEvent(self, e: Any) -> None:
        left_margin = 6
        if not self._cmap_is_full_width():
            left_margin = self._cmap_rect().right() + 4 - self.rect().x()
        self.setTextMargins(left_margin, 2, 0, 0)
        super().resizeEvent(e)

    def paintEvent(self, e: QPaintEvent) -> None:
        super().paintEvent(e)  # let the style paint background + text first

        cmap_rect = self._cmap_rect()
        if self._cmap:
            draw_colormap(
                self, self._cmap, cmap_rect, checkerboard_size=self._checkerboard_size
            )
        elif not self._cmap_is_full_width():
            if self._missing_cmap:
                draw_colormap(self, self._missing_cmap, cmap_rect)
            self._missing_icon.paint(QPainter(self), cmap_rect.adjusted(4, 4, 0, -4))



from __future__ import annotations

from typing import overload

from qtpy import QtCore, QtGui
from qtpy import QtWidgets as QtW
from qtpy.QtCore import Property, Qt


class QStyleOptionToggleSwitch(QtW.QStyleOptionButton):
    def __init__(self):
        super().__init__()
        self.on_color = QtGui.QColor("#4D79C7")
        self.off_color = QtGui.QColor("#909090")
        self.handle_color = QtGui.QColor("#d5d5d5")
        self.switch_width = 24
        self.switch_height = 12
        self.handle_size = 14

        # these aren't yet overridable in QToggleSwitch
        self.margin = 2
        self.text_alignment = Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter


class QToggleSwitch(QtW.QAbstractButton):
    StyleOption = QStyleOptionToggleSwitch

    @overload
    def __init__(self, parent: QtW.QWidget | None = ...) -> None: ...
    @overload
    def __init__(self, text: str | None, parent: QtW.QWidget | None = ...) -> None: ...

    def __init__(  # type: ignore [misc]  # overload
        self, text: str | None = None, parent: QtW.QWidget | None = None
    ) -> None:
        if isinstance(text, QtW.QWidget):
            if parent is not None:
                raise TypeError("No overload of QToggleSwitch matches the arguments")
            parent = text
            text = None

        # attributes for drawing the switch
        self._on_color = QtGui.QColor("#4D79C7")
        self._off_color = QtGui.QColor("#909090")
        self._handle_color = QtGui.QColor("#d5d5d5")
        self._switch_width = 24
        self._switch_height = 12
        self._handle_size = 14
        self._offset_value = 8.0

        super().__init__(parent)
        self.setCheckable(True)
        self.setCursor(Qt.CursorShape.PointingHandCursor)
        self.toggled.connect(self._animate_handle)

        self._anim = QtCore.QPropertyAnimation(self, b"_offset", self)
        self._anim.setDuration(120)
        self._offset_value = self._offset_for_checkstate(False)
        if text:
            self.setText(text)

    def initStyleOption(self, option: QStyleOptionToggleSwitch) -> None:
        """Initialize the style option for the switch."""
        option.initFrom(self)

        option.text = self.text()
        option.icon = self.icon()
        option.iconSize = self.iconSize()
        option.state |= (
            QtW.QStyle.StateFlag.State_On
            if self.isChecked()
            else QtW.QStyle.StateFlag.State_Off
        )

        option.on_color = self.onColor
        option.off_color = self.offColor
        option.handle_color = self.handleColor
        option.switch_width = self.switchWidth
        option.switch_height = self.switchHeight
        option.handle_size = self.handleSize

    def paintEvent(self, a0: QtGui.QPaintEvent | None) -> None:
        p = QtGui.QPainter(self)
        opt = QStyleOptionToggleSwitch()
        self.initStyleOption(opt)
        self.drawGroove(p, self._groove_rect(opt), opt)
        p.save()
        self.drawHandle(p, self._handle_rect(opt), opt)
        p.restore()
        self.drawText(p, self._text_rect(opt), opt)
        p.end()

    def minimumSizeHint(self) -> QtCore.QSize:
        return self.sizeHint()

    def setAnimationDuration(self, msec: int) -> None:
        """Set the duration of the animation in milliseconds.

        To disable animation, set duration to 0.
        """
        self._anim.setDuration(msec)

    def animationDuration(self) -> int:
        """Return the duration of the animation in milliseconds."""
        return self._anim.duration()

    def sizeHint(self) -> QtCore.QSize:
        self.ensurePolished()
        opt = QStyleOptionToggleSwitch()
        self.initStyleOption(opt)

        fm = QtGui.QFontMetrics(self.font())
        text_size = fm.size(0, self.text())
        height = max(opt.switch_height, text_size.height()) + opt.margin * 2
        width = opt.switch_width + text_size.width() + opt.margin * 2 + 8
        return QtCore.QSize(width, height)

    ### Re-implementable methods for drawing the switch ###

    def drawGroove(
        self,
        painter: QtGui.QPainter,
        rect: QtCore.QRectF,
        option: QStyleOptionToggleSwitch,
    ) -> None:
        """Draw the groove of the switch.

        Parameters
        ----------
        painter : QtGui.QPainter
            The painter to use for drawing.
        rect : QtCore.QRectF
            The rectangle in which to draw the groove.
        option : QStyleOptionToggleSwitch
            The style options used for drawing.
        """
        painter.setPen(Qt.PenStyle.NoPen)

        is_checked = option.state & QtW.QStyle.StateFlag.State_On
        is_enabled = option.state & QtW.QStyle.StateFlag.State_Enabled
        # draw the groove
        on_color = option.on_color
        off_color = option.off_color
        if is_enabled:
            painter.setBrush(on_color if is_checked else off_color)
            painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing, True)
            painter.setOpacity(0.8)
        else:
            if is_checked:
                ratio = 0.32
                mixed_color = QtGui.QColor(
                    int(on_color.red() * ratio + off_color.red() * (1 - ratio)),
                    int(on_color.green() * ratio + off_color.green() * (1 - ratio)),
                    int(on_color.blue() * ratio + off_color.blue() * (1 - ratio)),
                )
                painter.setBrush(mixed_color)
            else:
                painter.setBrush(off_color)
            painter.setOpacity(0.6)

        half_height = option.switch_height / 2
        painter.drawRoundedRect(rect, half_height, half_height)

    def drawHandle(
        self,
        painter: QtGui.QPainter,
        rect: QtCore.QRectF,
        option: QStyleOptionToggleSwitch,
    ) -> None:
        """Draw the handle of the switch.

        Parameters
        ----------
        painter : QtGui.QPainter
            The painter to use for drawing.
        rect : QtCore.QRectF
            The rectangle in which to draw the handle.
        option : QStyleOptionToggleSwitch
            The style options used for drawing.
        """
        painter.setPen(Qt.PenStyle.NoPen)
        painter.setBrush(option.handle_color)
        painter.setOpacity(1.0)
        painter.drawEllipse(rect)

    def drawText(
        self,
        painter: QtGui.QPainter,
        rect: QtCore.QRectF,
        option: QStyleOptionToggleSwitch,
    ) -> None:
        """Draw the text next to the switch.

        Parameters
        ----------
        painter : QtGui.QPainter
            The painter to use for drawing.
        rect : QtCore.QRectF
            The rectangle in which to draw the text.
        option : QStyleOptionToggleSwitch
            The style options used for drawing.
        """
        # TODO:
        # using self.style().drawControl(CE_PushButtonLabel ...)
        # might provide a more native experience.
        text_color = option.palette.color(self.foregroundRole())
        pen = QtGui.QPen(text_color, 1)
        painter.setPen(pen)
        painter.drawText(rect, int(option.text_alignment), option.text)

    ### Properties ###

    def _get_onColor(self) -> QtGui.QColor:
        return self._on_color

    def _set_onColor(self, color: QtGui.QColor | QtGui.QBrush) -> None:
        self._on_color = QtGui.QColor(color)
        self.update()

    onColor = Property(QtGui.QColor, _get_onColor, _set_onColor)
    """Color of the switch groove when it is on."""

    def _get_offColor(self) -> QtGui.QColor:
        return self._off_color

    def _set_offColor(self, color: QtGui.QColor | QtGui.QBrush) -> None:
        self._off_color = QtGui.QColor(color)
        self.update()

    offColor = Property(QtGui.QColor, _get_offColor, _set_offColor)
    """Color of the switch groove when it is off."""

    def _get_handleColor(self) -> QtGui.QColor:
        return self._handle_color

    def _set_handleColor(self, color: QtGui.QColor | QtGui.QBrush) -> None:
        self._handle_color = QtGui.QColor(color)
        self.update()

    handleColor = Property(QtGui.QColor, _get_handleColor, _set_handleColor)
    """Color of the switch handle."""

    def _get_switchWidth(self) -> int:
        return self._switch_width

    def _set_switchWidth(self, width: int) -> None:
        self._switch_width = width
        self._offset_value = self._offset_for_checkstate(self.isChecked())
        self.update()

    switchWidth = Property(int, _get_switchWidth, _set_switchWidth)
    """Width of the switch groove."""

    def _get_switchHeight(self) -> int:
        return self._switch_height

    def _set_switchHeight(self, height: int) -> None:
        self._switch_height = height
        self._offset_value = self._offset_for_checkstate(self.isChecked())
        self.update()

    switchHeight = Property(int, _get_switchHeight, _set_switchHeight)
    """Height of the switch groove."""

    def _get_handleSize(self) -> int:
        return self._handle_size

    def _set_handleSize(self, size: int) -> None:
        self._handle_size = size
        self.update()

    handleSize = Property(int, _get_handleSize, _set_handleSize)
    """Width/height of the switch handle."""

    ### Other private methods ###

    def _animate_handle(self, val: bool) -> None:
        end = self._offset_for_checkstate(val)
        if self._anim.duration():
            self._anim.setStartValue(self._offset_for_checkstate(not val))
            self._anim.setEndValue(end)
            self._anim.start()
        else:
            self._set_offset(end)

    def _get_offset(self) -> float:
        return self._offset_value

    def _set_offset(self, offset: float) -> None:
        self._offset_value = offset
        self.update()

    _offset = Property(float, _get_offset, _set_offset)

    def _offset_for_checkstate(self, val: bool) -> float:
        opt = QStyleOptionToggleSwitch()
        self.initStyleOption(opt)
        if val:
            offset = opt.margin + opt.switch_width - opt.switch_height / 2
        else:
            offset = opt.margin + opt.switch_height / 2
        return offset

    def _groove_rect(self, opt: QStyleOptionToggleSwitch) -> QtCore.QRectF:
        return QtCore.QRectF(
            opt.margin, self._vertical_offset(opt), opt.switch_width, opt.switch_height
        )

    def _handle_rect(self, opt: QStyleOptionToggleSwitch) -> QtCore.QRectF:
        return QtCore.QRectF(
            self._offset_value - opt.handle_size / 2,
            self._vertical_offset(opt) - (opt.handle_size - opt.switch_height) / 2,
            opt.handle_size,
            opt.handle_size,
        )

    def _text_rect(self, opt: QStyleOptionToggleSwitch) -> QtCore.QRectF:
        # If handle is bigger than groove, adjust the text to the right of the handle.
        # If groove is bigger, adjust the text to the right of the groove.
        return QtCore.QRectF(
            opt.switch_width
            + max(opt.handle_size - opt.switch_height, 0) // 2
            + opt.margin * 2
            + 2,
            0,
            self.width() - opt.switch_width - opt.margin * 2,
            self.height(),
        )

    def _vertical_offset(self, opt: QStyleOptionToggleSwitch) -> int:
        """Offset for the vertical centering of the switch."""
        return (self.height() - opt.switch_height) // 2 + opt.margin



import sys
from enum import Enum, EnumMeta, Flag
from functools import reduce
from itertools import combinations
from operator import or_
from typing import TypeVar

from qtpy.QtCore import Signal
from qtpy.QtWidgets import QComboBox

EnumType = TypeVar("EnumType", bound=Enum)


NONE_STRING = "----"


def _get_name(enum_value: Enum) -> str:
    """Create human readable name if user does not implement `__str__`."""
    str_module = getattr(enum_value.__str__, "__module__", "enum")
    if str_module != "enum" and not str_module.startswith("shibokensupport"):
        # check if function was overloaded
        name = str(enum_value)
    else:
        if enum_value.name is None:
            # This is hack for python bellow 3.11
            if not isinstance(enum_value, Flag):
                raise TypeError(
                    f"Expected Flag instance, got {enum_value}"
                )  # pragma: no cover
            if sys.version_info >= (3, 11):
                # There is a bug in some releases of Python 3.11 (for example 3.11.3)
                # that leads to wrong evaluation of or operation on Flag members
                # and produces numeric value without proper set name property.
                return f"{enum_value.value}"

            # Before python 3.11 there is no smart name set during
            # the creation of Flag members.
            # We needs to decompose the value to get the name.
            # It is under if condition because it uses private API.

            from enum import _decompose

            members, _not_covered = _decompose(enum_value.__class__, enum_value.value)
            name = "|".join(m.name.replace("_", " ") for m in members[::-1])
        else:
            name = enum_value.name.replace("_", " ")
    return name


def _get_name_with_value(enum_value: Enum) -> tuple[str, Enum]:
    return _get_name(enum_value), enum_value


class QEnumComboBox(QComboBox):
    """ComboBox presenting options from a python Enum.

    If the Enum class does not implement `__str__` then a human readable name
    is created from the name of the enum member, replacing underscores with spaces.
    """

    currentEnumChanged = Signal(object)

    def __init__(
        self, parent=None, enum_class: EnumMeta | None = None, allow_none=False
    ):
        super().__init__(parent)
        self._enum_class = None
        self._allow_none = False
        if enum_class is not None:
            self.setEnumClass(enum_class, allow_none)
        self.currentIndexChanged.connect(self._emit_signal)

    def setEnumClass(self, enum: EnumMeta | None, allow_none=False):
        """Set enum class from which members value should be selected."""
        self.clear()
        self._enum_class = enum
        self._allow_none = allow_none and enum is not None
        if allow_none:
            super().addItem(NONE_STRING)
        names_ = self._get_enum_member_list(enum)
        super().addItems(list(names_))

    @staticmethod
    def _get_enum_member_list(enum: EnumMeta | None):
        if issubclass(enum, Flag):
            members = list(enum.__members__.values())
            comb_list = []
            for i in range(len(members)):
                comb_list.extend(reduce(or_, x) for x in combinations(members, i + 1))

        else:
            comb_list = list(enum.__members__.values())
        return dict(map(_get_name_with_value, comb_list))

    def enumClass(self) -> EnumMeta | None:
        """Return current Enum class."""
        return self._enum_class

    def isOptional(self) -> bool:
        """Return if current enum is with optional annotation."""
        return self._allow_none

    def clear(self):
        self._enum_class = None
        self._allow_none = False
        super().clear()

    def currentEnum(self) -> EnumType | None:
        """Current value as Enum member."""
        if self._enum_class is not None:
            if self._allow_none:
                if self.currentText() == NONE_STRING:
                    return None
            return self._get_enum_member_list(self._enum_class)[self.currentText()]
        return None

    def setCurrentEnum(self, value: EnumType | None) -> None:
        """Set value with Enum."""
        if self._enum_class is None:
            raise RuntimeError(
                "Uninitialized enum class. Use `setEnumClass` before `setCurrentEnum`."
            )
        if value is None and self._allow_none:
            self.setCurrentIndex(0)
            return
        if not isinstance(value, self._enum_class):
            raise TypeError(
                "setValue(self, Enum): argument 1 has unexpected type "
                f"{type(value).__name__!r}"
            )
        self.setCurrentText(_get_name(value))

    def _emit_signal(self):
        if self._enum_class is not None:
            self.currentEnumChanged.emit(self.currentEnum())

    def insertItems(self, *_, **__):
        raise RuntimeError("EnumComboBox does not allow to insert items")

    def insertItem(self, *_, **__):
        raise RuntimeError("EnumComboBox does not allow to insert item")

    def addItems(self, *_, **__):
        raise RuntimeError("EnumComboBox does not allow to add items")

    def addItem(self, *_, **__):
        raise RuntimeError("EnumComboBox does not allow to add item")

    def setInsertPolicy(self, policy):
        raise RuntimeError("EnumComboBox does not allow to insert item")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast

from cmap import Colormap
from qtpy.QtCore import (
    QEvent,
    QObject,
    QSortFilterProxyModel,
    QStringListModel,
    Qt,
    Signal,
)
from qtpy.QtWidgets import (
    QButtonGroup,
    QCheckBox,
    QComboBox,
    QCompleter,
    QDialog,
    QDialogButtonBox,
    QMenu,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

from superqt.utils import signals_blocked

from ._catalog_combo import CmapCatalogComboBox
from ._cmap_item_delegate import QColormapItemDelegate
from ._cmap_line_edit import QColormapLineEdit
from ._cmap_utils import try_cast_colormap

if TYPE_CHECKING:
    from collections.abc import Sequence

    from cmap._colormap import ColorStopsLike
    from qtpy.QtGui import QKeyEvent, QMouseEvent


CMAP_ROLE = Qt.ItemDataRole.UserRole + 1


class QColormapComboBox(QComboBox):
    """A drop down menu for selecting colors.

    Parameters
    ----------
    parent : QWidget, optional
        The parent widget.
    allow_user_colormaps : bool, optional
        Whether the user can add custom colormaps by clicking the "Add
        Colormap..." item. Default is False. Can also be set with
        `setUserAdditionsAllowed`.
    add_colormap_text: str, optional
        The text to display for the "Add Colormap..." item.
        Default is "Add Colormap...".
    filterable: bool, optional
        Whether the user can filter colormaps by typing in the line edit.
        Default is True. Can also be set with `setFilterable`.
    """

    currentColormapChanged = Signal(Colormap)

    def __init__(
        self,
        parent: QWidget | None = None,
        *,
        allow_user_colormaps: bool = False,
        add_colormap_text: str = "Add Colormap...",
        filterable: bool = True,
    ) -> None:
        # init QComboBox
        super().__init__(parent)
        self._add_color_text: str = add_colormap_text
        self._allow_user_colors: bool = allow_user_colormaps
        self._last_cmap: Colormap | None = None
        self._filterable: bool = False

        line_edit = _PopupColormapLineEdit(self, allow_invalid=False)
        self.setLineEdit(line_edit)
        self.setInsertPolicy(QComboBox.InsertPolicy.NoInsert)
        self.setItemDelegate(QColormapItemDelegate(self))

        # there's a little bit of a potential bug here:
        # if the user clicks on the "Add Colormap..." item
        # then an indexChanged signal will be emitted, but it may not
        # actually represent a "true" change in the index if they dismiss the dialog
        self.activated.connect(self._on_activated)

        self.setUserAdditionsAllowed(allow_user_colormaps)

        # Create a proxy model to handle filtering
        self._proxy_model = QSortFilterProxyModel(self)
        # use string list model as source model
        self._proxy_model.setSourceModel(QStringListModel(self))
        self._proxy_model.setFilterCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)

        # Setup completer
        self._completer = QCompleter(self)
        self._completer.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        self._completer.setCompletionMode(QCompleter.CompletionMode.PopupCompletion)
        self._completer.setFilterMode(Qt.MatchFlag.MatchContains)
        self._completer.setModel(self._proxy_model)

        # set the delegate for both the popup and the combobox
        if popup := self._completer.popup():
            popup.setItemDelegate(self.itemDelegate())

        # Update completer model when items change
        if model := self.model():
            model.rowsInserted.connect(self._update_completer_model)
            model.rowsRemoved.connect(self._update_completer_model)

        self.setFilterable(filterable)

        self.currentIndexChanged.connect(self._on_index_changed)
        line_edit.editingFinished.connect(self._on_editing_finished)

        # Enable right-click "Remove" on dropdown items
        if (view := self.view()) and (viewport := view.viewport()):
            viewport.installEventFilter(self)

    def userAdditionsAllowed(self) -> bool:
        """Returns whether the user can add custom colors."""
        return self._allow_user_colors

    def setUserAdditionsAllowed(self, allow: bool) -> None:
        """Sets whether the user can add custom colors.

        If enabled, an "Add Colormap..." item will be added to the end of the
        list. When clicked, a dialog will be shown to allow the user to select
        a colormap from the
        [cmap catalog](https://cmap-docs.readthedocs.io/en/latest/catalog/).
        """
        self._allow_user_colors = bool(allow)

        idx = self.findData(self._add_color_text, Qt.ItemDataRole.DisplayRole)
        if idx < 0:
            if self._allow_user_colors:
                self.addItem(self._add_color_text)
        elif not self._allow_user_colors:
            self.removeItem(idx)

    def setFilterable(self, filterable: bool) -> None:
        """Set whether the user can enter/filter colormaps by typing in the line edit.

        If enabled, the user can select the text in the line edit and type to
        filter the list of colormaps. The completer will show a list of matching
        colormaps as the user types.  If disabled, the user can only select from
        the combo box dropdown.
        """
        self._filterable = bool(filterable)
        self.setCompleter(self._completer if self._filterable else None)
        self.lineEdit().setReadOnly(not self._filterable)

    def isFilterable(self) -> bool:
        """Returns whether the user can filter the list of colormaps."""
        return self._filterable

    def clear(self) -> None:
        super().clear()
        self.setUserAdditionsAllowed(self._allow_user_colors)
        self._update_completer_model()

    def itemColormap(self, index: int) -> Colormap | None:
        """Returns the color of the item at the given index."""
        return self.itemData(index, CMAP_ROLE)

    def addColormap(self, cmap: ColorStopsLike) -> None:
        """Adds the colormap to the QComboBox."""
        if (_cmap := try_cast_colormap(cmap)) is None:
            raise ValueError(f"Invalid colormap value: {cmap!r}")

        for i in range(self.count()):
            if item := self.itemColormap(i):
                if item.name == _cmap.name:
                    return  # no duplicates  # pragma: no cover

        had_items = self.count() > int(self._allow_user_colors)
        # add the new color and set the background color of that item
        self.addItem(_cmap.name.rsplit(":", 1)[-1])
        self.setItemData(self.count() - 1, _cmap, CMAP_ROLE)
        if not had_items:  # first item added
            self._on_index_changed(self.count() - 1)

        # make sure the "Add Colormap..." item is last
        idx = self.findData(self._add_color_text, Qt.ItemDataRole.DisplayRole)
        if idx >= 0:
            self._block_completer_update = True
            try:
                with signals_blocked(self):
                    self.removeItem(idx)
                    self.addItem(self._add_color_text)
            finally:
                self._block_completer_update = False

    def addColormaps(self, colors: Sequence[Any]) -> None:
        """Adds colors to the QComboBox."""
        self._block_completer_update = True
        try:
            for color in colors:
                self.addColormap(color)
        finally:
            self._block_completer_update = False
            self._update_completer_model()

    def currentColormap(self) -> Colormap | None:
        """Returns the currently selected Colormap or None if not yet selected."""
        return self.currentData(CMAP_ROLE)

    def setCurrentColormap(self, color: Any) -> None:
        """Adds the color to the QComboBox and selects it."""
        if not (cmap := try_cast_colormap(color)):
            raise ValueError(f"Invalid colormap value: {color!r}")

        for idx in range(self.count()):
            if (item := self.itemColormap(idx)) and item.name == cmap.name:
                # cmap_ is already here - just select it
                self.setCurrentIndex(idx)
                return

        # cmap not in the combo box yet — add it, then select it
        self.addColormap(cmap)
        idx = self.findData(cmap, CMAP_ROLE)
        if idx >= 0:
            old_idx = self.currentIndex()
            self.setCurrentIndex(idx)
            if idx == old_idx:
                # setCurrentIndex won't emit if the index didn't actually change
                # (e.g. first item added at index 0 when current index is already 0)
                self._on_index_changed(idx)

    def _on_activated(self, index: int) -> None:
        if self.itemText(index) != self._add_color_text:
            return

        dlg = _CmapNameDialog(self, Qt.WindowType.Sheet)
        if dlg.exec() and (cmap := dlg.combo.currentColormap()):
            # add the color and select it, without adding duplicates
            for i in range(self.count()):
                if (item := self.itemColormap(i)) and cmap.name == item.name:
                    self.setCurrentIndex(i)
                    return
            self.setCurrentColormap(cmap)
        elif self._last_cmap is not None:
            # user canceled, restore previous color without emitting signal
            idx = self.findData(self._last_cmap, CMAP_ROLE)
            if idx >= 0:
                with signals_blocked(self):
                    self.setCurrentIndex(idx)

    def _on_index_changed(self, index: int) -> None:
        colormap = self.itemData(index, CMAP_ROLE)
        if isinstance(colormap, Colormap):
            self.currentColormapChanged.emit(colormap)
            self.lineEdit().setColormap(colormap)
            self._last_cmap = colormap

    def _update_completer_model(self) -> None:
        """Update the completer's model with current items."""
        if getattr(self, "_block_completer_update", False):
            return

        # Ensure we are updating the source model of the proxy
        if isinstance(src_model := self._proxy_model.sourceModel(), QStringListModel):
            words = [
                txt
                for i in range(self.count())
                if (txt := self.itemText(i)) != self._add_color_text
            ]
            src_model.setStringList(words)
            self._proxy_model.invalidate()

    def _on_editing_finished(self) -> None:
        text = self.lineEdit().text()
        if (cmap := try_cast_colormap(text)) is not None:
            self.currentColormapChanged.emit(cmap)

            # if the cmap is not in the list, add it
            if self.findData(cmap, CMAP_ROLE) < 0:
                self.addColormap(cmap)

    def eventFilter(self, obj: QObject | None, event: QEvent | None) -> bool:
        if event and event.type() == QEvent.Type.MouseButtonRelease:
            mouse_event = cast("QMouseEvent", event)
            if mouse_event.button() == Qt.MouseButton.RightButton:
                view = self.view()
                if view and obj is view.viewport():
                    index = view.indexAt(mouse_event.pos())
                    if index.isValid():
                        text = self.itemText(index.row())
                        if text != self._add_color_text:
                            self._show_remove_menu(view, mouse_event.pos(), index.row())
                    return True  # consume the right-click
        return super().eventFilter(obj, event)

    def _show_remove_menu(self, view: Any, pos: Any, row: int) -> None:
        menu = QMenu(view)
        remove_action = menu.addAction("Remove")
        if menu.exec(view.viewport().mapToGlobal(pos)) == remove_action:
            was_current = row == self.currentIndex()
            self.removeItem(row)
            self._update_completer_model()
            if was_current:
                # select the first actual colormap, skipping "Add Colormap..."
                for i in range(self.count()):
                    if self.itemColormap(i) is not None:
                        self.setCurrentIndex(i)
                        return

    def keyPressEvent(self, e: QKeyEvent | None) -> None:
        if e and e.key() in (Qt.Key.Key_Enter, Qt.Key.Key_Return):
            # select the first completion when pressing enter if the popup is visible
            if (completer := self.completer()) and completer.completionCount():
                self.lineEdit().setText(completer.currentCompletion())  # type: ignore
        return super().keyPressEvent(e)


CATEGORIES = ("sequential", "diverging", "cyclic", "qualitative", "miscellaneous")


class _CmapNameDialog(QDialog):
    def __init__(self, *args: Any) -> None:
        super().__init__(*args)

        self.combo = CmapCatalogComboBox()

        B = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        btns = QDialogButtonBox(B)
        btns.accepted.connect(self.accept)
        btns.rejected.connect(self.reject)

        layout = QVBoxLayout(self)
        layout.addWidget(self.combo)

        self._btn_group = QButtonGroup(self)
        self._btn_group.setExclusive(False)
        for cat in CATEGORIES:
            box = QCheckBox(cat)
            self._btn_group.addButton(box)
            box.setChecked(True)
            box.toggled.connect(self._on_check_toggled)
            layout.addWidget(box)

        layout.addWidget(btns)
        self.setSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum)
        self.resize(self.sizeHint())

    def _on_check_toggled(self) -> None:
        # get valid names according to preferences
        word_list = Colormap.catalog().unique_keys(
            prefer_short_names=True,
            categories={b.text() for b in self._btn_group.buttons() if b.isChecked()},
        )
        self.combo.clear()
        self.combo.addItems(sorted(word_list))


class _PopupColormapLineEdit(QColormapLineEdit):
    def mouseReleaseEvent(self, _: Any) -> None:
        """Show parent popup when clicked.

        Without this, only the down arrow will show the popup.  And if mousePressEvent
        is used instead, the popup will show and then immediately hide.
        Also ensure that the popup is not shown when the user selects text.
        """
        if not self.hasSelectedText():
            parent = self.parent()
            if parent and hasattr(parent, "showPopup"):
                parent.showPopup()



from __future__ import annotations

import traceback
from typing import TYPE_CHECKING, cast

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QErrorMessage, QMessageBox, QWidget

if TYPE_CHECKING:
    from types import TracebackType


_DEFAULT_FLAGS = Qt.WindowType.Dialog | Qt.WindowType.MSWindowsFixedSizeDialogHint


class exceptions_as_dialog:
    """Context manager that shows a dialog when an exception is raised.

    See examples below for common usage patterns.

    To determine whether an exception was raised or not, check the `exception`
    attribute after the context manager has exited.  If `use_error_message` is `False`
    (the default), you can also access the `dialog` attribute to get/manipulate the
    `QMessageBox` instance.

    Parameters
    ----------
    exceptions : type[BaseException] | tuple[type[BaseException], ...], optional
        The exception(s) to catch, by default `Exception` (i.e. all exceptions).
    icon : QMessageBox.Icon, optional
        The icon to show in the QMessageBox, by default `QMessageBox.Icon.Critical`
    title : str, optional
        The title of the `QMessageBox`, by default `"An error occurred"`.
    msg_template : str, optional
        The message to show in the `QMessageBox`. The message will be formatted
        using three variables:

        - `exc_value`: the exception instance
        - `exc_type`: the exception type
        - `tb`: the traceback as a string

        The default template is the content of the exception: `"{exc_value}"`
    buttons : QMessageBox.StandardButton, optional
        The buttons to show in the `QMessageBox`, by default
        `QMessageBox.StandardButton.Ok`
    parent : QWidget | None, optional
        The parent widget of the `QMessageBox`, by default `None`
    use_error_message : bool | QErrorMessage, optional
        Whether to use a `QErrorMessage` instead of a `QMessageBox`. By default
        `False`. `QErrorMessage` shows a checkbox that the user can check to
        prevent seeing the message again (based on the text of the formatted
        `msg_template`.) If `True`, the global `QMessageError.qtHandler()`
        instance is used to maintain a history of dismissed messages. You may also pass
        a `QErrorMessage` instance to use a specific instance. If `use_error_message` is
        True, or if you pass your own `QErrorMessage` instance, the `parent` argument
        is ignored.

    Attributes
    ----------
    dialog : QMessageBox | None
        The `QMessageBox` instance that was created (if `use_error_message` was
        `False`).  This can be used, among other things, to determine the result of
        the dialog (e.g. `dialog.result()`) or to manipulate the dialog (e.g.
        `dialog.setDetailedText("some text")`).
    exception : BaseException | None
        Will hold the exception instance if an exception was raised and caught.

    Examples
    --------
    ```python
    from qtpy.QtWidgets import QApplication
    from superqt.utils import exceptions_as_dialog

    app = QApplication([])

    with exceptions_as_dialog() as ctx:
        raise Exception("This will be caught and shown in a QMessageBox")

    # you can access the exception instance here
    assert ctx.exception is not None

    # with exceptions_as_dialog(ValueError):
    #     1 / 0  # ZeroDivisionError is not caught, so this will raise

    with exceptions_as_dialog(msg_template="Error: {exc_value}"):
        raise Exception("This message will be inserted at 'exc_value'")

    for _i in range(3):
        with exceptions_as_dialog(AssertionError, use_error_message=True):
            assert False, "Uncheck the checkbox to ignore this in the future"

    # use ctx.dialog to get the result of the dialog
    btns = QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel
    with exceptions_as_dialog(buttons=btns) as ctx:
        raise Exception("This will be caught and shown in a QMessageBox")
    print(ctx.dialog.result())  # prints which button was clicked

    app.exec()  # needed only for the use_error_message example to show
    ```
    """

    dialog: QMessageBox | None
    exception: BaseException | None
    exec_result: int | None = None

    def __init__(
        self,
        exceptions: type[BaseException] | tuple[type[BaseException], ...] = Exception,
        icon: QMessageBox.Icon = QMessageBox.Icon.Critical,
        title: str = "An error occurred",
        msg_template: str = "{exc_value}",
        buttons: QMessageBox.StandardButton = QMessageBox.StandardButton.Ok,
        parent: QWidget | None = None,
        flags: Qt.WindowType = _DEFAULT_FLAGS,
        use_error_message: bool | QErrorMessage = False,
    ):
        self.exceptions = exceptions
        self.msg_template = msg_template
        self.exception = None
        self.dialog = None

        self._err_msg = use_error_message

        if not use_error_message:
            # the message will be overwritten in __exit__
            self.dialog = QMessageBox(
                icon, title, "An error occurred", buttons, parent, flags
            )

    def __enter__(self) -> exceptions_as_dialog:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        tb: TracebackType | None,
    ) -> bool:
        if not (exc_value is not None and isinstance(exc_value, self.exceptions)):
            return False  # let it propagate

        # save the exception for later
        self.exception = exc_value

        # format the message using the context variables
        if "{tb}" in self.msg_template:
            _tb = "\n".join(traceback.format_exception(exc_type, exc_value, tb))
        else:
            _tb = ""
        text = self.msg_template.format(exc_value=exc_value, exc_type=exc_type, tb=_tb)

        # show the dialog
        if self._err_msg:
            msg = (
                self._err_msg
                if isinstance(self._err_msg, QErrorMessage)
                else QErrorMessage.qtHandler()
            )
            cast("QErrorMessage", msg).showMessage(text)
        elif self.dialog is not None:  # it won't be if use_error_message=False
            self.dialog.setText(text)
            self.dialog.exec()

        return True  # swallow the exception



from collections.abc import Iterator
from contextlib import contextmanager
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from qtpy.QtCore import QObject


@contextmanager
def signals_blocked(*obj: "QObject") -> Iterator[None]:
    """Context manager to temporarily block signals emitted by QObjects.

    Parameters
    ----------
    *obj : QObject
        QObjects whose signals should be blocked.

    Examples
    --------
    ```python
    from qtpy.QtWidgets import QSpinBox
    from superqt import signals_blocked

    spinbox = QSpinBox()
    with signals_blocked(spinbox):
        spinbox.setValue(10)
    ```
    """
    previous = [o.blockSignals(True) for o in obj]
    try:
        yield
    finally:
        for o, prev in zip(obj, previous, strict=False):
            o.blockSignals(prev)
